#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

class Handler():
    def __init__(self):
        self.reset()

    def get_attr(self, attrs, attr):
        l = [v for k,v in attrs if k == attr]
        if l :
            return l[0]

class LinkHandler(Handler):
    def reset(self):
        self.active = 0
        self.link = ""
        self.content = ""
        self.handler = "link"

    def match(self, tag, attrs, open, ll):
        if tag == "a":
            if open:
                href = self.get_attr(attrs, "href")
                if href:
                    self.link = href
                    self.active = 1
                    return "%4"
                else:
                    self.reset()
            else:
                ll.append((self.content, self.link, self.handler))
                self.reset()
                return u"[" + unicode(len(ll)) + u"]%0"

class ImageHandler(Handler):
    def reset(self):
        self.active = 0
        self.handler = "image"

    def match(self, tag, attrs, open, ll):
        if tag == "img":
            if open:
                src = self.get_attr(attrs, "src")
                alt = self.get_attr(attrs, "alt")
                if not alt:
                    alt = "image"
                if src:
                    extension = src.rsplit('.',1)[-1]
                    ll.append((alt, src, self.handler))
                self.reset()
                return u"%7["+ alt + u"][" + unicode(len(ll)) + u"]%0"



# -*- coding: utf-8 -*-

#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from const import VERSION_TUPLE
import story
import tag

import cPickle
import fcntl

# Feed() controls a single feed and implements all of the update functionality
# on top of Tag() (which is the generic class for lists of items). Feed() is
# also the lowest granule for custom renderers because the renderers are
# most likely using information specific to the XML, rather than information
# specific to an arbitrary list.

# Each feed has a self.ufp item that contains a verbatim copy of the data
# returned by the feedparser.

# Each feed will also only write its status back to disk on tick() and only if
# has_changed() has been called by one of the Story() items Feed() contains.

class Feed(list):
    def __init__(self, cfg, dirpath, URL, tags, rate, keep, \
            renderer, filter, username, password):

        # Configuration set settings
        self.tags = tags
        self.base_set = 0

        self.URL = URL
        self.renderer = renderer
        self.rate = rate
        self.time = 1
        self.keep = keep
        self.username = username
        self.password = password

        # Hard filter
        if filter:
            self.filter = lambda x: filter(self, x)
        else:
            self.filter = None

        # Other necessities
        self.path = dirpath
        self.cfg = cfg
        self.changed = 0
        self.ufp = None
   
    def update(self):
        lockflags = fcntl.LOCK_SH
        if self.base_set:
            lockflags |= fcntl.LOCK_NB

        try:
            f = open(self.path, "r")
            try:
                fcntl.flock(f.fileno(), lockflags)
                self.ufp = cPickle.load(f)
            except:
                return 0
            finally:
                fcntl.flock(f.fileno(), fcntl.LOCK_UN)
                f.close()
        except:
            return 0

        # If this data pre-dates 0.6.0 (the last disk format update)
        # toss a key error.
        if "canto_version" not in self.ufp or\
                self.ufp["canto_version"][1] < 6:
            raise KeyError

        if not self.base_set:
            self.base_set = 1
            if "feed" in self.ufp and "title" in self.ufp["feed"]:
                replace = lambda x: x or self.ufp["feed"]["title"]
                self.tags = [ replace(x) for x in self.tags]
            else:
                # Using URL for tag, no guarantees
                self.tags = [self.URL] + self.tags

        self.extend(self.ufp["entries"])
        return 1

    def extend(self, entries):
        newlist = []
        for entry in entries:
            # If out todisk() lock failed, then it's possible
            # we have unwritten changes, so we need to move over
            # the canto_state, rather than just using the already
            # written one.

            selected = 0

            if entry in self:
                i = self.index(entry)
                if self.changed and entry["canto_state"] !=\
                        self[i]["canto_state"]:
                    entry["canto_state"] = self[i]["canto_state"]
                    self.has_changed()
                selected = self[i].sel

            # If tags were added in the configuration, c-f won't
            # notice (doesn't care about tags), so we check and
            # append as needed.

            for tag in self.tags:
                if tag not in entry["canto_state"]:
                    entry["canto_state"].append(tag)

            s = story.Story(entry, self, self.renderer)
            s.sel = selected
            newlist.append(s)

        del self[:]
        list.extend(self, filter(self.filter, newlist))

    def has_changed(self):
        self.changed = 1

    def todisk(self):
        f = open(self.path, "r+")
        try:
            fcntl.flock(f.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
            f.seek(0, 0)
            f.truncate()
            cPickle.dump(self.ufp, f)
            f.flush()
        except:
            return 0
        finally:
            fcntl.flock(f.fileno(), fcntl.LOCK_UN)
            f.close()

        self.changed = 0
        return 1

    def tick(self):
        if self.changed:
            self.todisk()

        self.time -= 1
        if self.time <= 0:
            if not self.update() or len(self) == 0 :
                self.time = 1
            else:
                self.time = self.rate



#!/usr/bin/python
# -*- coding: utf-8 -*-

#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from unittest import TestCase, main
import curses
import locale
import sys

width_override = 0

class TestCurses(TestCase):
    def setUp(self):
        self.screen = curses.initscr()
        self.screen.erase()
        self.height, self.width = self.screen.getmaxyx()
        if width_override:
            self.width = width_override
        curses.start_color()
        curses.init_pair(1, 7, 0)

        curses.noecho()
        curses.cbreak()
        self.screen.keypad(1)
        self.row = -1

    def tearDown(self):
        self.screen.refresh()
        self.screen.getch()
        curses.endwin()

from canto.interface_draw import Renderer
from canto.story import Story
from canto.tag import Tag
class TestIDraw(TestCurses):
    def setUp(self):
        TestCurses.setUp(self)
        for i,c in enumerate([7,4,3,2,5]):
            curses.init_pair(i + 1, c, 0)

        self.renderer = Renderer()
        self.row = 0
        
        self.ufp = {}
        self.ufp["title"] = u"Here, try an HTML entity on for size: &amp;"
        self.ufp["description"] =\
u"""Super long string description. With word wrapping and
line breaks
and slugified-huge-words-just-like-the-widecurses-test-has. Also
let's toss in some HTML, just for shits.
<p>
<pre><code>
    for(i=0;i &lt; n;i++) {
        //do something
        list[i] = list[i + 1];
    }
</code></pre>
</p>
<p>
<ol>
<li>Ordered 1</li>
<li>Ordered 2</li>
<li>Sublist <ul><li>Unordered 1</li><li>Unordered 2</li></ul>
<li>Ordered 4</li>
<li><strong>Styled Ordered 5</strong></li>
</ol>
<p>
<strong>Bold text</strong><br />
<em>Emphasis text</em>"""

        self.ufp["link"] = u"http://somelink"
        self.ufp["canto_state"] = [u"unread"]

        self.story = Story(self.ufp, None, self.renderer)
        self.story.idx = 0
        self.story.last = 1

        self.tag = Tag(None)

    def l(self, x):
        bme = ((""," ",""),(""," ",""),(""," ",""))
        self.row = self.renderer.out([(x, bme)], self.row, \
                self.height, self.width, [self.screen])
    
class TestIDrawBlocks(TestIDraw):
    def runTest(self):
        if self.width < 20:
            return

        self.width = 20
        # Simple quote
        self.l(u"%QThis is a quote%q")
        self.l(u" ")

        # Indent across a single line, that will break
        self.l(u"%IThis is uniformly indented broken line write.%i")
        self.l(u" ")

        # Multi line quote with forced breaks.
        self.l(u"%QThis is a")
        self.l(u"multi-line")
        self.l(u"quote%q")
        self.l(u" ")

class TestIDrawStory(TestIDraw):
    def runTest(self):
        self.renderer.story(None, self.tag, self.story, self.row, self.height,\
                self.width, [self.screen])

class TestIDrawReader(TestIDraw):
    def runTest(self):
        self.renderer.reader(None, self.story, self.width, 1, self.screen)

from canto.widecurse import core
class TestWidecurse(TestCurses):
    def l(self, x, r=" ", e=""):
        p = locale.getpreferredencoding()
        self.row += 1
        return core(self.screen, self.row, 0, self.width,\
                x.encode(p), 
                r.encode(p), 
                e.encode(p))

    def runTest(self):
        if self.width < 20:
            return

        self.width = 20
        # Terminal styles
        self.l(u"%8Normal output")
        self.l(u"%BBold output%b")
        self.l(u"%UUnderline output%u")
        self.l(u"%SStandout output%s")
        self.l(u"%RReverse output%r")
        self.l(u"%DDim output%d")
        
        # Color output
        for i in range(8):
            curses.init_pair(i + 1, i, 0)
            ouput = u"%" + unicode(i + 1) + unicode(i + 1) + u"%0"
            for off,attr in enumerate([u"", u"%B",u"%U",u"%S"]):
                core(self.screen, self.row, 0 + off, self.width, \
                        (attr + ouput +
                            attr.lower()).encode(locale.getpreferredencoding())\
                            , " ", "")
            self.row += 1

        # Basic Color layering
        self.l(u"%B%11%22%33%44%55%04%03%02%01%b")
        self.l(u"%8Cleanse the palette.")

        # Multi-line Color layering and end escapes
        self.l(u"%2 This is the color 2")
        self.l(u"This is color 2, too%0")
        self.l(u"But not this")

        # Empty string with repeating . ending in <><>
        self.l(u"", u".",u"<><>")

        # String of exact screen width
        self.l(u"12345678901234567890")

        # String of exact screen witdth, after escapes
        self.l(u"%21234567890%51234567890")

        # String too long
        r = self.l(u"%812345678901234567890extra")

        # Left over should be "extra"
        self.l(u"Left over: %s" % r)

        # Overflow the built-in color memory
        # This must be updated if color memory is increased
        self.l(u"%1z%2y%3x%4w%5v%6u%7t%8s%1r%2q%0a%0b%0c%0d%0e%0f%0g%0h%0i%0j")
        self.l(u" ")

        # Test word wrap
        s = u"%8A bunch of really short words, but more than one line."
        while s:
            s = self.l(s)

        self.l(u" ")

        # Test word breaking
        s = u"A-really-long-slugified-huge-word"
        while s:
            s = self.l(s)

from canto.canto_html import convert
class TestCantoHTML(TestCase):
    def runTest(self):
        # Test list nesting
        text, links = convert(u"<ul><li>Unordered</li><li>Some header text"
                u"<ol><li>Ordered</li><li>Also ordered</li></ol>"
                u"<li>Unordered, too</li></ul>")

        self.failUnless(text == u"\n%I\n\u25cf Unordered\n\u25cf "\
                u"Some header text\n%I\n1.Ordered\n2.Also ordered%i"\
                u"\n\n\u25cf Unordered, too%i\n")

        # Test LinkHandler
        text, links =  convert(u"""<a href="test">Blahblah</a>""")

        self.failUnless(links == [(u"%4Blahblah",u"test",u"link")])
        
        # Test ImageHandler
        text,links = convert(u"""<img src="myimage.jpg" />
                            <img src="otherimage.jpg" alt="Sexy" />""")

        self.failUnless(links == [(u"image",u"myimage.jpg",u"image"),\
                (u"Sexy",u"otherimage.jpg",u"image")])

if __name__ == "__main__":
    if len(sys.argv) > 1:
        try:
            width_override = int(sys.argv[1])
            sys.argv = sys.argv[:1]
        except:
            width_override = 0

    locale.setlocale(locale.LC_ALL, "")
    main()



# -*- coding: utf-8 -*-

#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

import interface_draw
import canto_html
import utility
import input

import subprocess
import locale
import time
import os
import re

# Adds Slashdot department information to reader
#
# Usage : add("Slashdot",\
#        "http://rss.slashdot.org/slashdot/Slashdot", \
#        renderer=slashdot_renderer()

class slashdot_renderer(interface_draw.Renderer):
    def reader_head(self, dict):
        title = self.do_regex(dict["story"]["title"], self.story_rgx)
        return [(u"%1%B" + title, u" ", u" "),\
                (u"%bfrom the " + dict["story"]["slash_department"] +\
                u" department%B", u" ", u" "),(u"┌",u"─",u"┐%C")]

# Adds a "tablist" to the default Canto config.

def tabbed_status(cfg):
    return u"%8%BCanto » %b%2" + \
            " ".join([unicode(x) for x in cfg.key_handlers]) + u"%1"

# Filter for filtering out all read stories.
#
# Usage : filters=[None, show_unread()]
#       then using [/] to cycle through.

class show_unread():
    def __str__(self):
        return "Show unread"

    def __call__(self, tag, item):
        return not item.wasread()

# Filter for filtering out all unread stories.
#
# Usage : filters=[None, show_marked()]
#       then using [/] to cycle through

class show_marked():
    def __str__(self):
        return "Show marked"

    def __call__(self, tag, item):
        return item.marked()

# A filter to take a keyword or regex and filter
# all stories that don't contain/match it.
#
# Usage : filters=[None, only_with("Obama")]
#         filters=[None, only_with(".*[Ll]inux.*", regex=True)]
#

class only_with():
    def __init__(self, keyword, **kwargs):
        self.keyword = keyword
        if "regex" in kwargs and kwargs["regex"]:
            self.match = re.compile(keyword)
        else:
            self.match = re.compile(".*" + keyword + ".*")

    def __str__(self):
        return "With %s" % self.keyword

    def __call__(self, tag, item):
        return self.match.match(item["title"])

# Same as above, except filters out all stories that
# *do* match the keyword / regex.

class only_without(only_with):
    def __str__(self):
        return "Without %s" % self.keyword

    def __call__(self, tag, item):
        return not self.match.match(item["title"])

# Display feed when it has one or more of the specified tags
class with_tag_in():
    def __init__(self, *tags):
        self.tags = set(tags)
    def __str__(self):
        return "With Tags: %s" % '/'.join(self.tags)

    def __call__(self, tag, item):
        tags=set(item.feed.tags)
        return bool(self.tags.intersection(tags))

# Display when all filters match
#
# Usage : filters=[all_of(with_tag_in('news'), show_unread)]

class all_of():
    def __init__(self, *filters):
        self.filters = [utility.get_instance(f) for f in filters]
    def __str__(self):
        return ' & '.join(["(%s)" % f for f in self.filters])

    def __call__(self, tag, item):
        return all([f(tag, item) for f in self.filters])

def set_filter(filter):
    filter = utility.get_instance(filter)
    return lambda x : x.set_filter(filter)

def set_tag_filter(filter):
    filter = utility.get_instance(filter)
    return lambda x : x.set_tag_filter(filter)

def set_tag_sort(sort):
    sort = utility.get_list_of_instances(sort)
    return lambda x : x.set_tag_sort(sort)

def set_tags(tags):
    return lambda x : x.set_tagset(tags)

# Creates a keybind for searching for a keyword or regex.
#
# Usage : keys["1"] = search("Obama")
#         keys["2"] = search(".*[Ll]inux.*, regex=True)

def search(s, **kwargs):
    if "regex" in kwargs and kwargs["regex"]:
        return lambda x : x.do_inline_search(re.compile(s))
    else:
        return lambda x : x.do_inline_search(re.compile(".*" + s + ".*"))

# Creates a keybind to do a interactive search.
#
# Usage : keys["/"] = search_filter

def search_filter(gui):
    rex = input.input(gui.cfg, "Search Filter")
    if not rex:
        return gui.set_filter(None)
    elif rex.startswith("rgx:"):
        rex = rex[4:]
    else:
        rex = "(?i).*" + re.escape(rex) + ".*"

    return gui.set_filter(only_with(rex, regex=True))

# Creates a keybind to append current story information to a file in the user's
# home directory. This is merely an example, but with a little modification it
# could be used to output XML chunks or Markdown output, etc.
# 
# Usage : keys["s"] = save

def save(x):
    file = open(os.getenv("HOME")+"/canto_out", "a")
    file.write(x.sel["title"] + "\n")
    file.write(x.sel["link"] + "\n\n")
    file.close()

# Creates a keybind to copy the URL of the current story to the clipboard.
# xclip must be available for this to work.
#
# Usage : keys["y"] = ["just_read", yank, "next_item"]

def yank(gui):
    xclip = subprocess.Popen('xclip -i', shell=True,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            stdin=subprocess.PIPE)
    try:
        xclip.stdin.write(gui.sel["link"])
        xclip.stdin.close()
        assert xclip.wait() == 0
    except (IOError, AssertionError):
        gui.cfg.log("xclip must be installed for yank to work!")
    else:
        gui.cfg.log("Yanked: %s" % gui.sel["title"])

# Note: the following two hacks are for xterm and compatible
# terminal emulators ([u]rxvt, eterm, aterm, etc.). These should
# not be run in screen or standard linux terms because they'll
# print garbage to the screen.

# Sets the xterm_title to Feed - Title
#
# Usage : select_hook = set_xterm_title

def set_xterm_title(tag, item):
    # Don't use print!
    prefcode = locale.getpreferredencoding()
    os.write(1, (u"\033]0; %s - %s\007" % \
            (tag.tag, item["title"])).encode(prefcode))

# Sets the xterm title to " "
#
# Usage : end_hook = clear_xterm_title

def clear_xterm_title(*args):
    os.write(1, "\033]0; \007")

# SORTS

class by_date:
    def __str__(self):
        return "By Date"

    def __call__(self, x, y):
        # We wrap this, despite the fact that sorts are all
        # wrapped in an exception logger because this is a
        # normal, unimportant problem.

        try:
            a = int(time.mktime(x["updated_parsed"]))
            b = int(time.mktime(y["updated_parsed"]))
        except:
            return 0

        return b - a

class by_len:
    def __str__(self):
        return "By Length"

    def __call__(self, x, y):
        return len(x["title"]) - len(y["title"])

class by_content:
    def __str__(self):
        return "By Length of Content"

    def __call__(self, x, y):
        def get_text(story):
            s,links = canto_html.convert(story.get_text())
            return s

        return len(get_text(x)) - len(get_text(y))

class by_alpha:
    def __str__(self):
        return "Alphabetical"

    def __call__(self, x, y):
        for a, b in zip(x["title"],y["title"]):
            if ord(a) != ord(b):
                return ord(a) - ord(b)

        return len(x["title"]) - len(y["title"])

class by_unread:
    def __str__(self):
        return "By Unread"

    def __call__(self, x, y):
        if x.wasread() and not y.wasread():
            return 1
        if y.wasread() and not x.wasread():
            return -1
        return 0

class reverse_sort:
    def __init__(self, other_sort):
        self.other_sort = utility.get_instance(other_sort)

    def __str__(self):
        return "Reversed %s" % self.other_sort

    def __call__(self, x, y):
        return -1 * self.other_sort(x,y)



# -*- coding: utf-8 -*-

#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from input import input, search, num_input
from const import *
import utility
import reader
import extra 

import curses

# Gui() is the class encompassing the basic view of canto,
# the list of feeds (tags) and items.

# Gui()'s main data structure is self.tags, which is a list
# of arbitrary Tag() objects, each being a list of stories.
# A corresponding list is self.map, which maps out the visible
# stories in order of appearance.

# Self.map may seem redundant, but it's mostly for convenience.
# For example, if the items are globally sorted, then iterating
# over self.tags won't work. Or if you're testing membership
# of a story in the self. list (see __select_topoftag).
# Or if your iterating over all visible items (see draw_elements)

# Self.list is still useful though, when dealing with things
# like tag based sorting, or setting the attributes of a Tag()
# since each story doesn't have access to its Tag() object
# directly.

class Gui :
    def __init__(self, cfg, list, tags, register, deregister):
        self.keys = cfg.key_list
        self.window_list = []
        self.map = []
        self.focus = 0

        self.cfg = cfg
        self.register = register
        self.deregister = deregister

        self.lines = 0
        self.sel = 0
        self.sel_idx = -1
        self.items = 0

        self.offset = 0
        self.max_offset = 0

        register(self)

        # Populate the Tag() objects provided with
        # stories from the list given.

        self.tags = tags
        for t in self.tags:
            t.extend(list)
        self.__do_new_hook()

        # Select the first visible feed.

        for t in self.tags :
            if len(t):
                self.sel = t[0]
                self.sel_idx = 0
                t[0].select()
                break
        else:
            self.cfg.log("No Items.")

        if self.cfg.start_hook:
            self.cfg.start_hook(self)

    def __str__(self):
        if self.focus:
            return "%B[base]%b"
        return "[base]"

    def refresh(self):
        # Generate all of the columns
        self.window_list = [curses.newwin(self.cfg.gui_height, \
                    self.cfg.gui_width / self.cfg.columns, \
                    self.cfg.gui_top,\
                    (self.cfg.gui_width / self.cfg.columns) * i\
                    + self.cfg.gui_right) \
                    for i in range(0, self.cfg.columns)]

        # Setup the backgrounds.
        for window in self.window_list:
            window.bkgdset(curses.color_pair(1))

        # Self.lines is the maximum number of visible lines on the screen
        # at any given time. Used for scroll detection.

        self.lines = self.cfg.columns * self.cfg.gui_height

        self.__map_items()
        self.draw_elements()

    def __map_items(self):

        # This for loop populates self.map with all stories that
        #   A - are first in a collapsed feed or not in one at all.
        #   B - that actually manage to print something to the screen.
        
        # Because of B, you can perform filtering in a Renderer().

        # We keep track of the virtual row to keep offsets in line.
        # It doesn't actually map to the row it's printed to on the
        # screen.

        self.map = []
        row = 0
        for i, feed in enumerate(self.tags):
            for item in feed:
                if not feed.collapsed or item.idx == 0:
                    item.lines = item.print_item(self.cfg, feed, 0, self)
                    if item.lines:
                        # item.tag_idx is the story's only reference
                        # to its current Tag()
                        item.tag_idx = i
                        item.row = row
                        row += item.lines
                        self.map.append(item)
        
        self.items = len(self.map)

        # Set max_offset, this is how we know not to recenter the
        # screen when it would leave unused space at the end.
        if self.items:
            self.max_offset = self.map[-1].row + \
                    self.map[-1].lines - self.lines

    def draw_elements(self):
        # Print all stories in self.map
        # Row increments always, because the drawing logic automatically
        # converts a row into a row in the proper window.

        if self.items > 0:
            self.__check_scroll()
            row = -1 * self.offset
            for item in self.map:
                # If row is not offscreen up
                if item.row + item.lines > self.offset:
                    # If row is offscreen down
                    if item.row > self.lines + self.offset:
                        break
                    item.print_item(self.cfg, self.tags[item.tag_idx], row, self)
                row += item.lines
        else:
            row = -1
        
        # Actually perform curses screen update.
        for i,win in enumerate(self.window_list) :
            if i * self.cfg.gui_height >= row:
                win.erase()
            else:
                win.clrtobot()
            win.noutrefresh()
        curses.doupdate()

    def __check_scroll(self) :
        # If our current item is offscreen up, ret 1
        if self.sel.row < self.offset :
            self.offset = self.sel.row
            return 1

        # If our current item is offscreen down, ret 1
        if self.sel.row + self.sel.lines > self.lines + self.offset :
            self.offset = self.sel.row + self.sel.lines - self.lines
            return 1
        return 0

    # This decorator makes items (usu. keybinds) that require
    # items to be present bail if none are.

    def noitem_unsafe(fn):
        def ns_dec(self, *args):
            if self.items > 0:
                return fn(self, *args)
            else:
                self.cfg.log("No Items.")
        return ns_dec

    # This decorator lets the bind just change sel_idx and
    # have self.sel set automatically and the story's state
    # synced.

    def change_selected(fn):
        def dec(self, *args):
            if self.sel_idx >= 0:
                if self.cfg.unselect_hook:
                    self.cfg.unselect_hook(self.tags[self.sel.tag_idx],
                            self.sel)
            oldsel = self.sel
            r = fn(self, *args)
            if oldsel:
                oldsel.unselect()
            if self.sel_idx >= 0:
                self.sel = self.map[self.sel_idx]
                self.sel.select()
                if self.cfg.select_hook:
                    self.cfg.select_hook(self.tags[self.sel.tag_idx], self.sel)
            return r
        return dec

    @change_selected
    def alarm(self, listobj):
        # Clear all of the tags and repopulate with the new listobj.
        # At this point, self.sel and self.sel_idx may be invalid

        for t in self.tags:
            t.clear()
            t.extend(listobj)

        self.__do_new_hook()
        self.__map_items() 

        # sel_idx may no longer be valid, because the item
        # list could shrink arbitrarily.
        self.sel_idx = min(self.sel_idx, self.items - 1)
        
        if self.items > 0:
            # Attempt to update sel_idx, if the item is still
            # visible (in self.map), otherwise just select
            # the top of the current (or first previous feed).

            if self.sel:
                # Since items can show up in multiple feeds
                # (i.e. reddit and subreddits), check that the
                # tag_idx is the same, so we don't jump from one
                # tag to another inadvertently.

                if self.sel in self.map and \
                        self.map[self.map.index(self.sel)].tag_idx == \
                        self.sel.tag_idx:
                    self.sel_idx = self.map.index(self.sel)
                else:
                    self.__select_topoftag()
            else:
                self.__select_topoftag(0)

        # If we had a selection, and now no items notify the user.
        elif self.sel:
            self.cfg.log("No Items.")
            self.sel = None

        if self.cfg.update_hook:
            self.cfg.update_hook(self)

    # Use the new_hook on any "new" items.
    # The new attribute is never accessible from the
    # renderer, and is only used for the hook.

    def __do_new_hook(self):
        if self.cfg.new_hook:
            for t in self.tags:
                for item in t:
                    if item.isnew():
                        self.cfg.new_hook(t, item)
                        item.old()

    def action(self, a):
        # Allows user defined functions to manipulate Gui()

        if hasattr(a, "__call__"):
            r = a(self)
        else:
            f = getattr(self, a, None)
            if f:
                r = f()
            else:
                r = -1

        if not r:
            self.draw_elements()
        return r

    @noitem_unsafe
    @change_selected
    def __select_topoftag(self, f=-1):
        if f < 0:
            f = self.sel.tag_idx
        for feed in self.tags[f:]:
            for item in feed:
                if item in self.map:
                    self.sel = item
                    self.sel_idx = self.map.index(self.sel)
                    return

    @change_selected
    def next_item(self):
        if self.sel_idx < self.items - 1:
            self.sel_idx += 1

    @change_selected
    def prev_item(self):
        if self.sel_idx > 0 :
            self.sel_idx -= 1

    @noitem_unsafe
    def prev_tag(self) :
        curtag = self.sel.tag_idx
        while not self.sel_idx == 0 :
            if curtag != self.sel.tag_idx and self.sel.idx == 0:
                break
            self.prev_item()

    @noitem_unsafe
    def next_tag(self) :
        curtag = self.sel.tag_idx
        while not self.sel_idx == self.items - 1:
            if curtag != self.sel.tag_idx:
                break
            self.next_item()

        # Next_tag should try to keep the top of the tag at
        # the top of the screen (as prev_tag does inherently)
        # so that the user's eye isn't lost.
        self.offset = min(self.sel.row, max(0, self.max_offset))

    # Goto_tagn goes to an absolute #'d tag. So the third
    # tag defined in your configuration will always be '3'

    @noitem_unsafe
    def goto_tag(self, num = None):
        if not num:
            num = num_input(self.cfg, "Absolute Tag")
        if not num:
            return

        if num < 0:
            num = len(self.tags) + num
        num = min(len(self.tags) - 1, num)

        while num > self.sel.tag_idx:
            self.next_tag()
    
        while num < self.sel.tag_idx:
            self.prev_tag()

    # Goto_reltagn goes to a tag relative to what's visible.

    @noitem_unsafe
    def goto_reltag(self, num = None):
        if not num:
            num = num_input(self.cfg, "Tag")
        if not num:
            return

        def rel_search(map):
            idx = -1
            cur = -1
            for item in map:
                if item.tag_idx != cur:
                    cur = item.tag_idx
                    idx += 1
                    if idx == num:
                        break
            return cur

        if num < 0:
            num = -1 * num - 1
            self.goto_tagn(rel_search(reversed(self.map)))
        else:
            self.goto_tagn(rel_search(self.map))

    @noitem_unsafe
    @change_selected
    def next_filtered(self, f) :
        cursor = self.sel_idx + 1
        while not cursor >= self.items:
            if f(self.tags[self.map[cursor].tag_idx],self.map[cursor]):
                self.sel_idx = cursor
                break
            cursor += 1

    @noitem_unsafe
    @change_selected
    def prev_filtered(self, f) :
        cursor = self.sel_idx - 1
        while not cursor < 0:
            if f(self.tags[self.map[cursor].tag_idx],self.map[cursor]):
                self.sel_idx = cursor
                break
            cursor -= 1

    def next_mark(self):
        self.next_filtered(extra.show_marked())

    def prev_mark(self):
        self.prev_filtered(extra.show_marked())

    def next_unread(self):
        self.next_filtered(extra.show_unread())

    def prev_unread(self):
        self.prev_filtered(extra.show_unread())

    def just_read(self):
        self.tags[self.sel.tag_idx].set_read(self.sel.idx)

    def just_unread(self):
        self.tags[self.sel.tag_idx].set_unread(self.sel.idx)

    @noitem_unsafe
    def goto(self) :        
        self.tags[self.sel.tag_idx].set_read(self.sel.idx)
        self.draw_elements()
        utility.goto(("", self.sel["link"], "link"), self.cfg)

    def help(self):
        self.cfg.wait_for_pid = utility.silentfork("man canto", "", 1, 0)

    @noitem_unsafe
    def reader(self) :
        self.tags[self.sel.tag_idx].set_read(self.sel.idx)
        reader.Reader(self.cfg, self.sel, self.register, self.deregister) 
        return REDRAW_ALL

    def change_filter(fn):
        def dec(self, *args):
            r,f = fn(self, *args)
            if r:
                self.cfg.log("Filter: %s" % f)
                return self.force_update(0)
        return dec

    @change_filter
    def set_filter(self, filt):
        return (self.cfg.filters.override(filt), self.cfg.filters.cur())

    @noitem_unsafe
    @change_filter
    def set_tag_filter(self, filt):
        return (self.tags[self.sel.tag_idx].filters.override(filt),\
                self.tags[self.sel.tag_idx].filters.cur())

    @change_filter
    def next_filter(self):
        return (self.cfg.filters.next(), self.cfg.filters.cur())
    
    @noitem_unsafe
    @change_filter
    def next_tag_filter(self):
        return (self.tags[self.sel.tag_idx].filters.next(),\
                self.tags[self.sel.tag_idx].filters.cur())

    @change_filter
    def prev_filter(self):
        return (self.cfg.filters.prev(), self.cfg.filters.cur())

    @noitem_unsafe
    @change_filter
    def prev_tag_filter(self):
        return (self.tags[self.sel.tag_idx].filters.prev(),\
                self.tags[self.sel.tag_idx].filters.cur())

    def change_sorts(fn):
        def dec(self, *args):
            r,s = fn(self, *args)
            if r:
                self.cfg.log("Sort: %s" % ", ".join([unicode(x) for x in s]))
                return ALARM
        return dec

    @change_sorts
    def next_tag_sort(self):
        return (self.tags[self.sel.tag_idx].sorts.next(),
                self.tags[self.sel.tag_idx].sorts.cur())

    @change_sorts
    def prev_tag_sort(self):
        return (self.tags[self.sel.tag_idx].sorts.prev(),
                self.tags[self.sel.tag_idx].sorts.cur())

    @change_sorts
    def set_tag_sort(self, sort):
        return (self.tags[self.sel.tag_idx].sorts.override(sort),\
                self.tags[self.sel.tag_idx].sorts.cur())

    def change_tags(fn):
        def dec(self, *args):
            r,t = fn(self, *args)
            if r:
                self.tags = t
                self.cfg.log("Tags: %s" % ", ".join(\
                        [unicode(x) for x in t if x]))
                return ALARM
        return dec

    @change_tags
    def next_tagset(self):
        return (self.cfg.tags.next(), self.cfg.tags.cur())

    @change_tags
    def prev_tagset(self):
        return (self.cfg.tags.prev(), self.cfg.tags.cur())

    @change_tags
    def set_tagset(self, t):
        return (1, self.cfg.get_real_tagl(t))

    @noitem_unsafe
    def inline_search(self):
        self.do_inline_search(search(self.cfg, "Inline Search"))

    def do_inline_search(self, s) :
        if s:
            for t in self.tags:
                for story in t:
                    if s.match(story["title"]):
                        story.mark()
                    else:
                        story.unmark()

            self.prev_mark()
            self.next_mark()
            self.draw_elements()

    @noitem_unsafe
    def toggle_mark(self):
        if self.sel.marked() :
            self.sel.unmark()
        else:
            self.sel.mark()

    @noitem_unsafe
    def all_unmarked(self):
        for item in self.map:
            if item.marked():
                item.unmark()

    @noitem_unsafe
    def toggle_collapse_tag(self):
        self.tags[self.sel.tag_idx].collapsed =\
                not self.tags[self.sel.tag_idx].collapsed
        self.sel.unselect()
        self.__map_items()
        self.__select_topoftag()

    def __collapse_all(self, c):
        for t in self.tags:
            t.collapsed = c
        self.__map_items()
        self.__select_topoftag()

    def set_collapse_all(self):
        self.__collapse_all(1)

    def unset_collapse_all(self):
        self.__collapse_all(0)

    def force_update(self, v = 1):
        if v:
            self.cfg.log("Forcing update.")

        for f in self.cfg.feeds :
            f.time = 1
        return ALARM
    
    @noitem_unsafe
    def tag_read(self):
        self.tags[self.sel.tag_idx].all_read()

    def all_read(self):
        for t in self.tags:
            t.all_read()

    @noitem_unsafe
    def tag_unread(self):
        self.tags[self.sel.tag_idx].all_unread()

    def all_unread(self):
        for t in self.tags :
            t.all_unread()

    def switch(self):
        return WINDOW_SWITCH

    def quit(self):
        if self.cfg.end_hook:
            self.cfg.end_hook(self)
        self.deregister()
        return -1



# -*- coding: utf-8 -*-

#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from curses import ascii
import curses
import signal
import re

# I am aware that Python's curses library comes with a TextBox class
# and, indeed, the input() function was using it for awhile. The problems
# with Textbox were numerous though:
#   * Only ASCII characters could be printed/inputted (the *big* one)
#   * Included a whole bunch of multi-line editing and validation stuff
#       that was completely unnecessary, since we know the input line
#       only needs to be one line long.
#   * To make editing easier, it used a half-ass system of gathering
#       the data from the window's written content with win.inch(),
#       which apparently didn't play nice with multi-byte?
#
# All of these problems have been fixed in half as many lines with all
# the same functionality on a single line basis, but the design is still
# based on Textbox.

class InputBox:
    def __init__(self, win):
        self.minx = win.getyx()[1]
        self.x = self.minx

        self.win = win
        self.win.keypad(1)
        self.result = ""

    def refresh(self):
        self.win.move(0, self.minx)
        maxx = self.win.getmaxyx()[1]
        try:
            self.win.addstr(self.result[-1 * (maxx - self.minx):]\
                    .encode("UTF-8", "replace"))
        except:
            pass
        self.win.clrtoeol()
        self.win.move(0, min(self.x, maxx - 1))
        self.win.refresh()

    def key(self, ch):
        if ch in (ascii.STX, curses.KEY_LEFT):
            if self.x > self.minx:
                self.x -= 1
        elif ch in (ascii.BS, curses.KEY_BACKSPACE):
            if self.x > self.minx:
                idx = self.x - self.minx
                self.result = self.result[:idx - 1] + self.result[idx:]
                self.x -= 1
        elif ch in (ascii.ACK, curses.KEY_RIGHT): # C-f
            self.x += 1
            if len(self.result) + self.minx < self.x:
                self.result += " "
        elif ch in (ascii.ENQ, curses.KEY_END): # C-e
            self.x = self.minx + len(self.result)
        elif ch in (ascii.SOH, curses.KEY_HOME): # C-a
            self.x = self.minx
        elif ch == ascii.NL: # C-j
            return 0
        elif ch == ascii.BEL: # C-g
            return -1
        elif ch == ascii.FF: # C-l
            self.refresh()
        else:
            self.x += 1
            idx = self.x - self.minx
            self.result = self.result[:idx] + unichr(ch) + self.result[idx:]
        return 1

    def edit(self):
        while 1:
            ch = self.win.getch()
            if ch <= 0:
                continue
            r = self.key(ch)
            if not r:
                break
            if r < 0:
                self.result = None
                break
            self.refresh()
        return self.result

def input(cfg, prompt):
    cfg.message("%B%1" + prompt + ":%b ")

    cfg.msg.move(0, len(prompt) + 2)

    temp = signal.getsignal(signal.SIGALRM)
    signal.signal(signal.SIGALRM, signal.SIG_IGN)
 
    curses.curs_set(1)
    term = InputBox(cfg.msg).edit()
    curses.curs_set(0)

    signal.signal(signal.SIGALRM, temp)
    signal.alarm(1)

    cfg.msg.erase()
    cfg.msg.refresh()

    return term

def num_input(cfg, prompt):
    term = input(cfg, prompt)
    if not term:
        return

    try:
        term = int(term)
    except:
        cfg.log("Not a number.")
        return None

    return term

def search(cfg, prompt):
    term = input(cfg, prompt)
    if not term :
        return

    elif term.startswith("rgx:"):
        str = term[4:]
    else:
        str = ".*" + re.escape(term) + ".*"
    
    try:
        m = re.compile(str)
    except:
        return None

    return m



#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from const import VERSION_TUPLE

from threading import Thread
import feedparser
import commands
import urlparse
import urllib2
import cPickle
import socket
import signal
import shutil
import fcntl
import time
import sys
import os

def main(cfg, optlist, verbose=False, force=False):

    socket.setdefaulttimeout(30)

    threads = []

    # Because canto-fetch isn't an ncurses application,
    # we might actually want to print to the screen!

    def log_func(x):
        if verbose:
            print x
        cfg.log(x)

    def imdone():
        if threads != []:
            for thread in threads:
                thread.join()
        log_func("Gracefully exiting Canto-fetch.")
        return 1

    killme = lambda a, b: imdone() and sys.exit(0)

    signal.signal(signal.SIGTERM, killme)
    signal.signal(signal.SIGINT, killme)

    for opt,arg in optlist:
        if opt in ["-V","--verbose"]:
            verbose = True
        elif opt in ["-f","--force"]:
            force = True

    # Make sure that the feed_dir does, indeed, exist and is
    # actually a directory.

    if not os.path.exists(cfg.feed_dir):
        os.mkdir(cfg.feed_dir)
    elif not os.path.isdir(cfg.feed_dir):
        os.unlink(cfg.feed_dir)
        os.mkdir(cfg.feed_dir)

    # Remove any crap out of the directory. This is mostly for
    # cleaning up when the user has removed a feed from the configuration.

    valid_names = [f.URL.replace("/"," ") for f in cfg.feeds]
    for file in os.listdir(cfg.feed_dir):
        if not file in valid_names:
            log_func("Deleted extraneous file: %s" % file)
            try:
                os.unlink(cfg.feed_dir + file)
            except:
                pass

    # The main canto-fetch loop.
    for fd in cfg.feeds:
        fpath = cfg.feed_dir + fd.URL.replace("/", " ")
        spath = cfg.script_dir
        threads.append(UpdateThread(fd, fpath, spath, force, log_func))
        threads[-1].start()

    imdone()
    return 0

class UpdateThread(Thread):
    def __init__(self, fd, fpath, spath, force, log_func):
        Thread.__init__(self)
        self.fd = fd
        self.fpath = fpath
        self.spath = spath
        self.force = force
        self.log_func = log_func
        self.prevtime = 0

        # This emptyfeed forms a skeleton for any canto feed.
        # Canto_state is a place holder. Canto_update is the
        # last time the feed was updated, and canto_version is
        # obviously the version under which it was last written.

        # Canto_version will be used in the future, if later
        # releases change anything serious in the format.

        self.emptyfeed = {"canto_state":[], "entries":[], "canto_update":0,
                        "canto_version":VERSION_TUPLE}

    def get_curfeed(self):
        curfeed = self.emptyfeed
        if os.path.exists(self.fpath) and os.path.isfile(self.fpath):
            f = open(self.fpath, "r")
            fcntl.flock(f.fileno(), fcntl.LOCK_SH)

            self.prevtime = os.stat(self.fpath).st_mtime

            try:
                curfeed = cPickle.load(f)
            except:
                self.log_func("cPickle load exception on %s" % self.fpath)
            finally:
                fcntl.flock(f.fileno(), fcntl.LOCK_UN)
                f.close()

        return curfeed

    # Now we attempt to load the previous feed information.

    def run(self):
        curfeed = self.get_curfeed()

        # Determine whether it's been long enough between
        # updates to warrant refetching the feed.

        if time.time() - curfeed["canto_update"] < self.fd.rate * 60 and\
                not self.force:
            return

        # Attempt to set the tag, if unspecified, by grabbing
        # it out of the previously downloaded info.

        if not self.fd.base_set:
            if "feed" in curfeed and "title" in curfeed["feed"]:
                replace = lambda x: x or curfeed["feed"]["title"]
                self.fd.tags = [ replace(x) for x in self.fd.tags]
                self.fd.base_set = 1
                self.log_func("Updating %s" % self.fd.tags[0])
            else:
                # This is the first time we've gotten this URL,
                # so just use the URL since we don't know the title.

                self.log_func("New feed %s" % self.fd.URL)
        else:
            self.log_func("Updating %s" % self.fd.tags[0])

        try:
            if self.fd.URL.startswith("script:"):
                script = self.spath + "/" + self.fd.URL[7:]
                out = commands.getoutput(script)
                newfeed = feedparser.parse(out)
            else:
                request = urllib2.Request(self.fd.URL)
                request.add_header('User-Agent',\
                    "Canto/%d.%d.%d + http://codezen.org/canto" %\
                    VERSION_TUPLE)

                if self.fd.username or self.fd.password:
                    mgr = urllib2.HTTPPasswordMgrWithDefaultRealm()
                    domain = urlparse.urlparse(self.fd.URL)[1]
                    mgr.add_password(None, domain,\
                            self.fd.username, self.fd.password)

                    # First, we try Basic Authentication
                    auth = urllib2.HTTPBasicAuthHandler(mgr)
                    opener = urllib2.build_opener(auth)
                    try:
                        newfeed = feedparser.parse(opener.open(request))
                    except:
                        # And, failing that, try Digest Authentication
                        auth = urllib2.HTTPDigestAuthHandler(mgr)
                        opener = urllib2.build_opener(auth)
                        newfeed = feedparser.parse(opener.open(request))
                else:
                    newfeed = feedparser.parse(\
                            feedparser.urllib2.urlopen(request))
        except:
            # Generally an exception is a connection refusal, but in any
            # case we either won't get data or can't trust the data, so
            # just skip processing this feed.

            self.log_func("Exception trying to get feed %s : %s" % \
                    (self.fd.tags[0], sys.exc_info()[1]))
            return

        # I don't know why feedparser doesn't actually throw this
        # since all URLErrors are basically unrecoverable.

        if "bozo_exception" in newfeed:
            if type(newfeed["bozo_exception"]) == urllib2.URLError:
                self.log_func(\
                    "Feedparser exception getting %s : %s, bailing." %\
                    (self.fd.URL, newfeed["bozo_exception"].reason))
                return
            if not len(newfeed["entries"]):
                self.log_func(\
                    "Feedparser exception, no content in %s : %s, bailing." %\
                    (self.fd.URL, newfeed["bozo_exception"]))
                return

        if not self.fd.base_set:
            if "feed" not in newfeed or "title" not in newfeed["feed"]:
                self.log_func("Ugh. Defaulting to URL for tag. No guarantees.")
                newfeed["feed"]["title"] = self.fd.URL

            replace = lambda x: x or newfeed["feed"]["title"]
            self.fd.tags = [ replace(x) for x in self.fd.tags]

        # Feedparser returns a very nice dict of information.
        # if there was something wrong with the feed (usu. encodings
        # being mis-declared or missing tags), it sets
        # bozo_exception.

        # These exceptions are recoverable and their objects are
        # un-Picklable so we log it and remove the value.

        if "bozo_exception" in newfeed:
            self.log_func("Recoverable error in feed %s: %s" % 
                        (self.fd.tags[0], newfeed["bozo_exception"]))
            newfeed["bozo_exception"] = None

        # Make state persist between feeds
        newfeed["canto_state"] = curfeed["canto_state"]
        newfeed["canto_update"] = time.time()

        # We can set this here, without checking curfeed.
        # Any migration should be done in the get_curfeed function,
        # when the old data is first loaded.

        newfeed["canto_version"] = VERSION_TUPLE

        # For all content that we would usually use, we convert
        # it to UTF-8 and escape all %s with \. Feedparser
        # almost without exception gives us all string in Unicode
        # so none of these should fail.

        def escape(s):
            s = s.replace("\\","\\\\")
            return s.replace("%", "\\%")

        for key in newfeed["feed"]:
            if type(newfeed["feed"][key]) in [unicode,str]:
                newfeed["feed"][key] = escape(newfeed["feed"][key])

        for entry in newfeed["entries"]:
            for subitem in ["content","enclosures"]:
                if subitem in entry:
                    for e in entry[subitem]:
                        for k in e.keys():
                            if type(e[k]) in [unicode,str]:
                                e[k] = escape(e[k])

            for key in entry.keys():
                if type(entry[key]) in [unicode,str]:
                    entry[key] = escape(entry[key])

        for entry in newfeed["entries"]:
            # If the item didn't come with a GUID, then
            # use link and then title as an identifier.

            if not "id" in entry:
                if "link" in entry:
                    entry["id"] = entry["link"]
                elif "title" in entry:
                    entry["id"] = entry["title"]
                else:
                    entry["id"] = None

        # Then search through the current feed to
        # make item state persistent, and loop until
        # it's safe to update on disk.

        while 1:
            for entry in newfeed["entries"]:
                for centry in curfeed["entries"]:
                    if entry["id"] == centry["id"]:
                        entry["canto_state"] = centry["canto_state"]
                        
                        # The entry is removed so that later it's
                        # not a candidate for being appended to the
                        # end of the feed.

                        curfeed["entries"].remove(centry)
                        break

                # Apply default state to genuinely new items.
                if not "canto_state" in entry:
                    entry["canto_state"] = self.fd.tags + \
                            [u"*", u"new"]

            # Tailor the list to the correct number of items.
            if len(newfeed["entries"]) < self.fd.keep:
                newfeed["entries"] += \
                    curfeed["entries"][:self.fd.keep - len(newfeed["entries"])]
            else:
                newfeed["entries"] = newfeed["entries"][:self.fd.keep]

            # Dump the output to the new file.
            f = open(self.fpath, "a")
            fcntl.flock(f.fileno(), fcntl.LOCK_EX)

            # The feed was modified out from under us.
            if self.prevtime and self.prevtime != os.stat(self.fpath).st_mtime:
                fcntl.flock(f.fileno(), fcntl.LOCK_UN)
                f.close()

                newer_curfeed = self.get_curfeed()

                # There was an actual c-f update done, bail
                if newer_curfeed["canto_update"] != curfeed["canto_update"]:
                    self.log_func("%s updated already, bailing" %
                            self.fd.tags[0])
                    break

                # Just a state modification by the client, update and continue.
                else:
                    curfeed = newer_curfeed
                    continue

            f.seek(0, 0)
            f.truncate()
            try:
                cPickle.dump(newfeed, f)
                f.flush()
            except:
                self.log_func("cPickle dump exception on %s" % self.fpath)
            finally:
                fcntl.flock(f.fileno(), fcntl.LOCK_UN)
                f.close()

            break




# -*- coding: utf-8 -*-

#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

import tempfile
import urllib2
import signal 
import curses
import sys
import re
import os

# The Cycle class has proved to be useful. It's used
# to encapsulate every cycle in canto, global filters,
# tag filters, global and tag sorts, tags. It's
# essentially a list with a current pointer and
# exception proof next/prev function and the ability
# to temporarily override a particular value.

class Cycle():
    def __init__(self, list, idx = 0):
        self.over = None
        self.list = list
        if 0 <= idx < len(self.list):
            self.idx = idx
        else:
            self.idx = 0

    def next(self):
        self.over = None
        if self.idx >= len(self.list) - 1:
            return 0
        self.idx += 1
        return 1

    def prev(self):
        self.over = None
        if self.idx <= 0:
            return 0
        self.idx -= 1
        return 1

    def override(self, cur):
        if self.over != cur:
            self.over = cur
            return 1
        return 0

    def cur(self):
        return self.over or self.list[self.idx]

# The get_instance() and get_list_of_instances()
# functions are to ensure that all objects (usu.
# in a Cfg()) are actual, instantiated objects and
# not just references.

def get_instance(l):
    if not l:
        return l
    return (hasattr(l, "__class__") and l) or l()

def get_list_of_instances(l):
    if not hasattr(l, "__iter__"):
        l = [l]

    r = []
    for i in l:
        if hasattr(i, "__iter__"):
            r.append(get_list_of_instances(i))
        else:
            r.append(get_instance(i))
    return r

def daemonize():
    pid = os.fork()
    if not pid:
        # New terminal session
        os.setsid()

        os.chdir("/")
        os.umask(0)
        pid = os.fork()
        if pid:
            sys.exit(0)
    else:
        sys.exit(0)

    # Close all possible terminal output
    # file descriptors. 

    os.close(0)
    os.close(1)
    os.close(2)

def convcolor(c):
    colordir = {"default" : -1, 
            "black" : 0, 
            "white" : 7, 
            "red" : 1, 
            "green" : 2, 
            "yellow" : 3, 
            "blue" : 4, 
            "magenta" : 5, 
            "pink" : 5, 
            "cyan" : 6}

    if type(c) == int:
        if 0 <= c <= curses.COLORS:
            return c
        else:
            return 0
    elif type(c) == str:
        if c in colordir:
            return colordir[c]
    return 0

def convkey(s):
    if len(s) == 1:
        return (ord(s),0)
    elif s.startswith("C-"):
        k, m = convkey(s[2:])
        
        # & 0x1F indicates CTRL status.
        return (k & 0x1F, m) 

    elif s.startswith("M-"):
        k, m = self.convkey(s[2:])
        return (k, 1)

    #For some reason, RETURN isn't in curses
    elif s == "KEY_RETURN":
        return (10, 0)
    else:
        return (getattr(curses, s), 0)

def conv_key_list(dict):
    ret = {}
    for key in dict:
        if not dict[key]:
            continue

        try:
            newkey = convkey(key)
        except AttributeError:
            continue

        # All the items in a key_list must be lists.
        if type(dict[key]) != type([]):
            ret[newkey] = [dict[key]]
        else:
            ret[newkey] = dict[key]

    return ret

def silentfork(path, href, text, fetch):

    pid = os.fork()
    if not pid :
        # A lot of programs don't appreciate
        # having their fds closed, so instead
        # we dup them to /dev/null.

        fd = os.open("/dev/null", os.O_RDWR)
        os.dup2(fd, sys.stderr.fileno())

        if not text:
            os.setpgid(os.getpid(), os.getpid())
            os.dup2(fd, sys.stdout.fileno())

        if fetch:
            response = urllib2.urlopen(href)
            data = response.read()
            fd, name = tempfile.mkstemp()
            os.write(fd, data)
            os.close(fd)
            path = path.replace("%u", name)
        else:
            path = path.replace("%u", href)

        os.system(path)
        sys.exit(0)

    if text:
        signal.signal(signal.SIGALRM, signal.SIG_IGN)
        signal.signal(signal.SIGWINCH, signal.SIG_IGN)

    return pid

def goto(link, cfg):
    title,href,handler = link
    if handler in cfg.handlers:
        for k in [h for h in cfg.handlers[handler].keys() if h]:
            if href.endswith(k):
                binary, text, fetch = cfg.handlers[handler][k]
                break
        else:
            if None in cfg.handlers[handler]:
                binary, text, fetch = cfg.handlers[handler][None]
            else:
                cfg.log("No default %s handler defined!" % handler)
                return


        # Escape all "s in the URL, to avoid malicious use
        # of crafted feeds. Thanks to Andreas.
        href = href.replace("\"","%22")

        if text:
            cfg.wait_for_pid = silentfork(binary, href, 1, fetch)
        else:
            silentfork(binary, href, 0, fetch)
    else:
        cfg.log("No handler set for %s" % handler)

def stripchars(string):
    string = string.replace("\\","\\\\")
    string = string.replace("%", "\\%")
    return string

def strip_escape_chars(strings):
    return (re.sub("\\\\(.)", "\\1", string) for string in strings)



# -*- coding: utf-8 -*-

#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

UPDATE_FIRST = 1
CHECK_NEW = 2
FEED_LIST = 4
OUT_OPML = 8
IN_OPML = 16
IN_URL = 32

REFRESH_ALL = 1
READER_NEXT = 2
READER_PREV = 3
ALARM = 4
KEY_PASSTHRU = 5
REDRAW_ALL = 6
WINDOW_SWITCH = 7

VERSION_TUPLE = SET_VERSION_TUPLE



# -*- coding: utf-8 -*-

#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from input import num_input
from const import *
import utility

import curses

class Reader :
    def __init__(self, cfg, story, register, deregister):
        self.story = story
        self.cfg = cfg
        self.keys = cfg.reader_key_list
        self.focus = 0
        self.more = 0
        self.offset = 0
        self.height = 0
        self.width = 0
        self.height = 0
        self.show_links = 0

        register(self)
        self.register = register
        self.deregister = deregister
        self.refresh()

    def __str__(self):
        if self.focus:
            return u"%B[" + self.story["title"][:10] + u"]%b"
        return u"[" + self.story["title"][:10] + u"]"

    def refresh(self):
        # It's unfortunate, but because the interface is so complex,
        # the only way to get the number of lines it will take to completely
        # render the reader, we actually have to render it to a None window
        # first.

        # A way to get this right off the bat would be nice, but I doubt
        # it would enhance the performance more than one iota.

        if self.cfg.reader_orientation in ["top","bottom",None]:
            # First render for self.lines
            self.lines, self.links = self.story.renderer.reader(self.cfg, \
                    self.story, self.cfg.width, self.show_links, None)

            # This is the default, old behavior (floating window)
            if not self.cfg.reader_orientation:
                self.height, self.width = min(self.lines, self.cfg.gui_height),\
                        self.cfg.width
                self.top, self.right = (0,0)
            # Rendering the reader into a pre-existing space
            else:
                self.height = self.cfg.reader_lines
                self.width = self.cfg.width
                if self.cfg.reader_orientation == "top":
                    self.top, self.right = (0,0)
                else:
                    self.top, self.right = (self.cfg.gui_height, 0)
        else:
            self.lines, self.links = self.story.renderer.reader(self.cfg, \
                    self.story, self.cfg.reader_lines, self.show_links, None)

            self.height = self.cfg.gui_height
            self.width = self.cfg.reader_lines

            if self.cfg.reader_orientation == "left":
                self.top, self.right = (0, 0)
            else:
                self.top, self.right = (0, self.cfg.gui_width)
                
        self.window = curses.newpad(self.lines, self.width)
        self.window.bkgdset(curses.color_pair(1))
        self.lines, self.links = self.story.renderer.reader(self.cfg, self.story, \
                self.width, self.show_links, self.window)

        self.draw_elements()

    def draw_elements(self):
        self.more = self.lines - (self.height + self.offset)
        self.window.refresh(self.offset, 0, self.top, self.right, \
                self.height - 1 + self.top, self.width + self.right)

    def toggle_show_links(self):
        self.show_links = not self.show_links
        self.refresh()

        if not self.show_links:
            return REDRAW_ALL

    def scroll_down(self):
        if self.more > 0 :
            self.offset += 1

    def page_down(self):
        if self.more > self.height:
            self.offset += self.height
        else:
            self.offset = self.lines - self.height

    def scroll_up(self):
        if self.offset :
            self.offset -= 1
    
    def page_up(self):
        if self.offset > self.height:
            self.offset -= self.height
        else:
            self.offset = 0

    def goto(self):
        self.dogoto(num_input(self.cfg, u"Link Number"))

    def dogoto(self, n):
        if n == None:
            return
        if n < len(self.links):
            utility.goto(self.links[n], self.cfg)
        return 1

    def switch(self):
        return WINDOW_SWITCH

    def alarm(self, a=None, b=None):
        pass
    
    def action(self, a):
        if hasattr(a, "__call__"):
            r = a(self)
        else:
            r = 0
            f = getattr(self,a,None)
            if f:
                r = f()

        if not r:
            self.draw_elements()
        return r

    def quit(self):
        self.destroy()
        return REDRAW_ALL

    def destroy(self):
        self.window.erase()
        self.draw_elements()
        self.deregister()



#!/usr/bin/python
# -*- coding: utf-8 -*-

#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from unittest import TestCase, main
import canto.cfg

class TestCfg(TestCase):
    def setUp(self):
        self.c = canto.cfg.Cfg("/dev/null", "/dev/null", "/dev/null",
        "/dev/null")

class TestHookExceptions(TestCfg):
    def runTest(self):
        self.c.parse(\
"""
def bad_resize_hook(cfg):
    tmp = 1 / 0
resize_hook = bad_resize_hook
""")

        self.c.resize_hook(None)

class TestInstantiation(TestCfg):
    def runTest(self):
        self.c.parse(\
"""
from canto.extra import *
# Feed.renderer should be instantiated
add("someURL", renderer=renderer)

# So should default tag sorts and filters
default_tag_sorts([by_unread, [None, by_len()]])
default_tag_filters([None, show_unread(), show_marked])

# And global filters
filters=[ None, show_unread(), show_marked]

# And custom tags
add_tag("Sometag", sorts=[ by_unread, [None, by_len()]])
""")
        self.failUnless(hasattr(self.c.feeds[0].renderer, "__class__"))

        for s in [self.c.tag_sorts, self.c.cfgtags[0].sorts.list]:
            self.failUnless(hasattr(s[0], "__class__"))
            self.failUnless(hasattr(s[1][1], "__class__"))

        for l in [self.c.tag_filters, self.c.filters.list]:
            for i in l:
                self.failUnless((not i) or hasattr(i, "__class__"))

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from const import VERSION_TUPLE
from utility import Cycle
import interface_draw
import utility
import feed
import tag

import xml.parsers.expat
import traceback
import chardet
import codecs
import os

class Cfg:
    def __init__(self, conf, log_file, feed_dir, script_dir):
        self.handlers = {
            "link" : {},
            "image" : {}
        }

        self.wait_for_pid = 0
        self.log_file = log_file

        # A MYRIAD of defaults. Bwahaha.

        self.key_list = {"q" : "quit",
                         "KEY_DOWN" : "next_item",
                         "KEY_UP" : "prev_item",
                         "j" : "next_item",
                         "k" : "prev_item",
                         "KEY_RIGHT" : "just_read",
                         "KEY_LEFT" : "just_unread",
                         "KEY_NPAGE" : "next_tag",
                         "KEY_PPAGE" : "prev_tag",
                         "[" : "prev_filter",
                         "]" : "next_filter",
                         "{" : "prev_tag_filter",
                         "}" : "next_tag_filter",
                         "-" : "prev_tag_sort",
                         "=" : "next_tag_sort",
                         "l" : "next_tag",
                         "o" : "prev_tag",
                         "<" : "prev_tagset",
                         ">" : "next_tagset",
                         "g" : "goto",
                         "." : "next_unread",
                         "," : "prev_unread",
                         "f" : "inline_search",
                         "n" : "next_mark",
                         "p" : "prev_mark",
                         " " : "reader",
                         "c" : "toggle_collapse_tag",
                         "C" : "set_collapse_all",
                         "V" : "unset_collapse_all",
                         "m" : "toggle_mark",
                         "M" : "all_unmarked",
                         "r" : "tag_read",
                         "R" : "all_read",
                         "u" : "tag_unread",
                         "U" : "all_unread",
                         ";" : "goto_reltag",
                         ":" : "goto_tag",
                         "C-r" : "force_update",
                         "C-l" : "refresh",
                         "\t" : "switch",
                         "h" : "help"}
        
        self.reader_key_list = {"KEY_DOWN" : "scroll_down",
                              "KEY_UP" : "scroll_up",
                              "j" : "scoll_down",
                              "k" : "scroll_up",
                              "KEY_NPAGE" : "page_down",
                              "KEY_PPAGE" : "page_up",
                              "g" : "goto",
                              "l" : "toggle_show_links",
                              "n" : ["destroy","next_item","reader"],
                              "p" : ["destroy","prev_item","reader"],
                              "h" : ["destroy","help"],
                              "q" : ["destroy","quit"],
                              "\t" : "switch",
                              " " : "quit"}

        self.colors = [("white","black"),("blue","black"),("yellow","black"),
                ("green","black"),("pink","black"),("black","black"),("blue","black"),(0,0)]

        self.feeds = []
        self.tags = [None]
        self.cfgtags = []

        self.default_rate = 5
        self.default_keep = 40
        self.default_filter = None
        self.default_sort = None
        self.default_renderer = interface_draw.Renderer()
        self.default_msg_tick = 5

        # Key handlers is a stack-like list that contains all "inputs"
        # that can take keys from the user. Generally, this is every
        # graphical class open at a time. The last item being the top
        # window, receiving keys.

        self.key_handlers = []
        self.cur_kh = -1

        self.path = conf
        self.feed_dir = feed_dir
        self.script_dir = script_dir

        self.columns = 1
        self.height = 0
        self.width = 0

        self.gui_top = 0
        self.gui_right = 0
        self.gui_height = 0
        self.gui_width = 0

        self.msg_height = 1
        self.msg = None
        self.msg_tick = 0

        self.status = default_status

        self.reader_lines = 0
        self.reader_orientation = None

        self.resize_hook = None
        self.new_hook = None
        self.select_hook = None
        self.unselect_hook = None
        self.start_hook = None
        self.end_hook = None
        self.update_hook = None

        self.tag_filters = [None]
        self.tag_sorts = [[None]]
        self.filters = [None]

        self.no_conf = 0

        # If we can't stat self.path, generate a default config
        # and toss a message about making your own.

        try :
            os.stat(self.path)
        except :
            print "Unable to find config file. Generating and "\
                  "using ~/.canto/conf.example"
            print "You will keep getting this until you create your "\
                  "own ~/.canto/conf"
            print "\nRemember: it's 'h' for help.\n"

            newpath = os.getenv("HOME") + "/.canto/"
            if not os.path.exists(newpath):
                os.mkdir(newpath)

            self.path = newpath + "conf.example"
            f = codecs.open(self.path, "w", "UTF-8")
            f.write("# Auto-generated by canto because you don't have one.\n"
                    "# Please copy to/create ~/.canto/conf\n\n")
            f.write("""add("""\
                    """"http://rss.slashdot.org/slashdot/Slashdot")\n""")
            f.write("""add("""\
                    """"http://reddit.com/.rss")\n""")
            f.write("""add("""\
                    """"http://kerneltrap.org/node/feed")\n""")
            f.write("""add("""\
                    """"http://codezen.org/canto/feeds/latest")\n""")
            f.write("\n")
            f.close()
            self.no_conf = 1

    def message(self, s, time=0):
        if self.msg:
            self.default_renderer.status(self.msg, self.msg_height,\
                    self.width, s)
            if not time:
                self.msg_tick = self.default_msg_tick
            else:
                self.msg_tick = time
            self.msg.refresh()

    # Simple append log.

    def log(self, message, mode="a"):
        self.message(message)
        try:
            f = open(self.log_file, mode)
            f.write(message + "\n")
            f.close()
        except:
            pass

    def wrap_args(self, kwargs):
        for attr in ["renderer", "filter"]:
            if attr in kwargs:
                kwargs[attr] = utility.get_instance(kwargs[attr])
        if "filter" in kwargs:
            kwargs["filter"] = self.filter_dec(kwargs["filter"])

        return kwargs

    def addfeed(self, tag, URL, **kwargs):
        print "'add_feed' is deprecated! Update to new 'add' syntax."
        print "See codezen.org/canto for instructions."
        kwargs["tags"] = [tag]
        self.add(URL, **kwargs)

    def add(self, URL, **kwargs):
        if (not URL) or URL == "":
            return -1

        for key in ["keep","rate","renderer"]:
            if not key in kwargs:
                kwargs[key] = getattr(self, "default_" + key)

        for key in ["username","password", "filter"]:
            if not key in kwargs:
                kwargs[key] = None

        if not "tags" in kwargs:
            kwargs["tags"] = [None]
        else:
            tgs = []
            for tag in kwargs["tags"]:
                if tag:
                    if type(tag) != unicode:
                        tgs.append(unicode(tag, "UTF-8", "ignore"))
                    else:
                        tgs.append(tag)
                else:
                    tgs.append(None)
            kwargs["tags"] = tgs

        kwargs = self.wrap_args(kwargs)

        # The tag is the only thing that has to be unique, so we ignore
        # any duplicate URLs, or everything  will break.

        if not URL in [f.URL for f in self.feeds]:
            self.feeds.append(feed.Feed(self, self.feed_dir +\
                    URL.replace("/", " "), URL,
                    kwargs["tags"],
                    kwargs["rate"],
                    kwargs["keep"],
                    kwargs["renderer"],
                    kwargs["filter"],
                    kwargs["username"],
                    kwargs["password"]))
            return 1
        return -1

    def set_default_rate(self, rate):
        self.default_rate = rate

    def set_default_keep(self, keep):
        self.default_keep = keep

    def set_default_renderer(self, renderer):
        self.default_renderer = renderer

    def set_default_tag_filters(self, filters):
        self.tag_filters = utility.get_list_of_instances(filters)

    def set_default_tag_sorts(self, sorts):
        self.tag_sorts = utility.get_list_of_instances(sorts)

    def change_feed(self, URL, **kwargs):
        l = [f for f in self.feeds if f.URL == URL]
        if not len(l):
            return

        kwargs = self.wrap_args(kwargs)

        feed = l[0]
        for key in ["keep","rate","renderer","filter","username","password"]:
            if key in kwargs:
                setattr(feed, key, kwargs[key])

    # This decorator-like function is used to wrap
    # all of the hooks, so that user exceptions don't
    # take Canto down.

    def hook_dec(self, fn):
        if not fn:
            return None

        def hdec(*args):
            try:
                r = fn(*args)
            except:
                self.log("\nException in hook:")
                self.log("%s" % traceback.format_exc())
                return 0
            return r
        return hdec

    def filter_dec(self, c):
        if not c:
            return None

        class fdec():
            def __init__(self, instance, log):
                self.instance = instance
                self.log = log

            def __str__ (self):
                return self.instance.__str__()

            def __call__(self, *args):
                try:
                    return self.instance(*args)
                except:
                    self.log("\nException in filter:")
                    self.log("%s" % traceback.format_exc())
        return fdec(c, self.log)

    def read_decode(self, filename, top_encode=0):
        enc = "utf-8"

        try:
            f = open(filename, "r")
            data = f.read()

            try:
                ret = unicode(data, enc)
            except UnicodeDecodeError:
                # If the Python built-in decoders can't figure it
                # out, it might need some help from chardet.
                enc = chardet.detect(data)["encoding"]
                ret = unicode(data, enc)
                self.log("Chardet detected encoding %s for %s" %\
                        (enc,filename))
        except :
            self.log("Failed to open config! (%s)" % sys.exc_info())
        finally:
            f.close()

        if top_encode and not ret.startswith("# -*- coding:"):
            ret = "# -*- coding: " + enc + " -*-\n" + ret

        return ret

    def parse(self, data = None):

        locals = {"addfeed":self.addfeed,
            "add_feed": self.addfeed,
            "add": self.add,
            "add_tag" : self.add_tag,
            "tags" : self.tags,
            "change_feed": self.change_feed,
            "default_rate" : self.set_default_rate,
            "default_keep" : self.set_default_keep,
            "default_renderer" : self.set_default_renderer,
            "default_tag_filters" : self.set_default_tag_filters,
            "default_tag_sorts" : self.set_default_tag_sorts,
            "renderer" : interface_draw.Renderer,
            "status" : self.status,
            "keys" : self.key_list,
            "reader_keys" : self.reader_key_list,
            "reader_orientation" : self.reader_orientation,
            "reader_lines" : self.reader_lines,
            "columns" : self.columns,
            "colors" : self.colors,
            "source_opml" : self.source_opml,
            "source_urls" : self.source_urls,
            "link_handler" : self.link_handler,
            "image_handler" : self.image_handler}

        # The entirety of the config is read in first (rather
        # than using execfile) because the config could be in
        # some strange encoding, and execfile would choke attempting
        # to coerce some character into ASCII.

        if not data:
            data = self.read_decode(self.path, 1)

        try :
            exec(data.encode("UTF-8"), {}, locals)
        except :
            print "Invalid line in config."
            traceback.print_exc()
            raise

        # exec cannot modify basic type
        # locals directly, so we do it by hand.

        for attr in ["render", "columns", "reader_orientation",\
                "reader_lines", "status", "tags"]:
            if attr in locals:
                setattr(self, attr, locals[attr])

        for attr in ["filters"]:
            if attr in locals:
                setattr(self, attr, \
                        utility.get_list_of_instances(locals[attr]))

        # Deprecated attributes... will be eliminated > 0.6.0

        for attr in ["browser","text_browser"]:
            if attr in locals:
                print "Use of %s is deprecated! Use link_handler() instead." %\
                    attr
                if attr == "browser":
                    self.handlers["link"][None] = (locals[attr], 0, 0)
                elif attr == "text_browser":
                    self.handlers["link"][None] =\
                            (self.handlers["link"][None][0], 1, 0)

        # Wrap hooks in the exception handler
        for hook in ["resize_hook","new_hook","select_hook","update_hook",\
                "unselect_hook","start_hook","end_hook"]:
            if hook in locals:
                setattr(self, hook, self.hook_dec(locals[hook]))

        # Wrap filters in exception handler
        self.filters = Cycle([self.filter_dec(x) for x in self.filters])
        self.tag_filters = [self.filter_dec(x) for x in self.tag_filters]

        # Ensure we have at least one column
        if not self.columns:
            self.log("columns <1, set to 1")
            self.columns = 1

        # Convert all of the C-M-blah (human readable) keys into
        # key tuples used in the main loop.

        self.key_list = utility.conv_key_list(self.key_list)
        self.reader_key_list = utility.conv_key_list(self.reader_key_list)

    def source(fn):
        def source_dec(self, *args, **kwargs):
            append = False
            if "append" in kwargs:
                append = kwargs["append"]
                file = codecs.open(self.path, "a", "UTF-8")

            l = fn(self, *args, **kwargs)

            for f in l:
                if self.add(f[0], tags=[f[1]]) and append:
                    if f[1]:
                        file.write(u"""add("%s", tags=["%s"])\n""" % f)
                    else:
                        file.write(u"""add("%s")\n""" % f[0])

            if append:
                file.close()
        return source_dec
   
    @source
    def source_opml(self, filename, **kwargs):
        l = []
        def start(name, attrs) : 
            if name == "outline" and (\
                (("type" in attrs and\
                attrs["type"] in ["pie","rss"])) or\
                not ("type" in attrs)):

                if "xmlUrl" in attrs:
                    if "text" in attrs:
                        l.append((attrs["xmlUrl"], attrs["text"]))
                    else:
                        l.append((attrs["xmlUrl"], None))

        p = xml.parsers.expat.ParserCreate()
        p.StartElementHandler = start
        d = self.read_decode(filename)
        p.Parse(d.encode("UTF-8"), 1)
        return l

    @source
    def source_urls(self, filename, **kwargs):
        l = []
        d = self.read_decode(filename).split('\n')[:-1]
        for feed in d:
            l.append((feed, None))
        return l

    @source
    def source_url(self, URL, **kwargs):
        if "tag" in kwargs:
            return [(URL, kwargs["tag"])]
        return [(None, URL)]

    def handler(self, handlers, path, **kwargs):
        if not "text" in kwargs:
            kwargs["text"] = False
        if not "fetch" in kwargs:
            kwargs["fetch"] = False
        if not "ext" in kwargs:
            kwargs["ext"] = None
        handlers.update(\
                {kwargs["ext"] : (path, kwargs["text"], kwargs["fetch"])})

    def image_handler(self, path, **kwargs):
        self.handler(self.handlers["image"], path, **kwargs)

    def link_handler(self, path, **kwargs):
        self.handler(self.handlers["link"], path, **kwargs)

    def add_tag(self, tags, **kwargs):
        if "sorts" in kwargs:
            kwargs["sorts"] = \
                Cycle(utility.get_list_of_instances(kwargs["sorts"]))
        else:
            kwargs["sorts"] = Cycle([[None]])

        if "filters" in kwargs:
            kwargs["filters"] = \
                    Cycle(utility.get_list_of_instances(kwargs["filters"]))
        else:
            kwargs["filters"] = Cycle(self.tag_filters)

        if not hasattr(tag, "__iter__"):
            tags = [tags]

        for t in tags:
            self.cfgtags.append(tag.Tag(\
                    self,
                    kwargs["sorts"],
                    kwargs["filters"],
                    unicode(t, "UTF-8", "ignore")))

    def get_real_tagl(self, tl):
        if not tl:
            tl = [ f.tags[0] for f in self.feeds ]
        if not hasattr(tl, "__iter__"):
            tl = [tl]

        r = []
        for t in tl:
            if t and type(t) != unicode:
                t = unicode(t, "UTF-8", "ignore")
            newtag = tag.Tag(self, Cycle(self.tag_sorts),\
                    Cycle(self.tag_filters), t)

            if newtag in self.cfgtags:
                newtag = self.cfgtags[self.cfgtags.index(newtag)]
            r.append(newtag)

        return r

    def validate_tags(self):
        # Change tags into actual tag objects
        self.tags = Cycle([ self.get_real_tagl(x) for x in self.tags ])

def default_status(cfg):
    return u"%8%B" + u"Canto %d.%d.%d" % VERSION_TUPLE + u"%b%1"



# -*- coding: utf-8 -*-

#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

# Story() controls a single story and is always contained in a feed().
# Like Feed(), it contains a self.ufp variable that holds a verbatim copy of
# the data returned by the UFP. The class then mostly serves as a soft wrapper
# around that data.

# Data that is persistent (everything except being selected) is appended to
# the canto_state key in self.ufp and is written to disk by the Feed() object
# automatically.

class Story():
    def __init__(self, ufp, feed, renderer):
        self.tag_idx = 0
        self.idx = 0
        self.last = 0

        self.row = 0
        self.lines = 0
        
        self.ufp = ufp
        self.feed = feed
        self.sel = 0
        self.renderer = renderer

    def __eq__(self, other):
        if self["id"] != other["id"]:
            return 0
        return 1

    def __getitem__(self, key):
        if key in self.ufp:
            return self.ufp[key]
        else:
            return ""

    def __contains__(self, key):
        return key in self.ufp

    def tagwrap(self, tag, i):
        if i == 0:
            return tag in self.ufp["canto_state"]
        elif i == 1 and not tag in self.ufp["canto_state"]:
            self.ufp["canto_state"].append(tag)
        elif i == -1 and tag in self.ufp["canto_state"]:
            self.ufp["canto_state"].remove(tag)

        if self.feed:
            self.feed.has_changed()

    def wasread(self):
        return self.tagwrap("read", 0)

    def read(self):
        self.tagwrap("read", 1)

    def unread(self):
        self.tagwrap("read", -1)

    def marked(self):
        return self.tagwrap("marked", 0)

    def mark(self):
        self.tagwrap("marked", 1)

    def unmark(self):
        self.tagwrap("marked", -1)

    def isnew(self):
        return self.tagwrap("new", 0)

    def new(self):
        self.tagwrap("new", 1)

    def old(self):
        self.tagwrap("new", -1)

    def selected(self):
        return self.sel

    def select(self):
        self.sel = 1

    def unselect(self):
        self.sel = 0

    def print_item(self, cfg, tag, row, i):
        return self.renderer.story(cfg, tag, self, row, \
                i.cfg.gui_height, i.cfg.gui_width / i.cfg.columns, \
                i.window_list)

    def get_text(self):
        if "content" in self:
            for c in self["content"]:
                if "text" in c["type"]:
                    return c["value"]

        return self["description"]



#!/usr/bin/env python

#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

#   This was inspired by Aaron Swartz's html2text, but doesn't do
#   file IO, doesn't do markdown, and doesn't shy away from Unicode.

from handlers import LinkHandler, ImageHandler
from HTMLParser import HTMLParser
import htmlentitydefs
import re

class CantoHTML(HTMLParser):

    # Reset is used, instead of __init__ so a single
    # instance of the class can parse multiple HTML
    # fragments.

    def reset(self):
        HTMLParser.reset(self)
        self.result = ""
        self.list_stack = []
        self.verbatim = 0

        self.links = []
        self.mime_handlers = [LinkHandler(),ImageHandler()]

    # unknown_* funnel all tags to handle_tag

    def handle_starttag(self, tag, attrs):
        self.handle_tag(tag, attrs, 1)

    def handle_endtag(self, tag):
        self.handle_tag(tag, {}, 0)

    def handle_data(self, text):
        if self.verbatim <= 0:
            text = text.replace(u"\n", u" ")

        for handler in self.mime_handlers:
            if handler.active:
                handler.content += text

        self.result += text

    # convert_* are called by SGMLParser's default
    # handle_char/entityref functions.

    def convert_charref(self, ref):
        try:
            if ref[0] in [u'x',u'X']:
                c = int(ref[1:], 16)
            else:
                c = int(ref)
        except:
            return u"[?]"
        return unichr(c)

    def handle_charref(self, ref):
        self.result += self.convert_charref(ref)

    def convert_entityref(self, ref):
        if ref in htmlentitydefs.name2codepoint:
            return unichr(htmlentitydefs.name2codepoint[ref])
        return u"[?]"

    def handle_entityref(self, ref):
        self.result += self.convert_entityref(ref)

    # This is the real workhorse of the HTML parser.

    def handle_tag(self, tag, attrs, open):
        for handler in self.mime_handlers:
            output = handler.match(tag, attrs, open, self.links)
            if output:
                self.handle_data(output)

        if tag in [u"h" + unicode(x) for x in xrange(1,7)]:
            if open:
                self.result += u"\n%B"
            else:
                self.result += u"%b\n"
        if tag in [u"blockquote"]:
            if open:
                self.result += u"\n%Q"
            else:
                self.result += u"%q\n"
        elif tag in [u"pre",u"code"]:
            if open:
                if tag == u"pre":
                    self.result += u"\n%Q"
                self.verbatim += 1
            else:
                if tag == u"pre":
                    self.result += u"%q\n"
                self.verbatim -= 1
        elif tag in [u"sup"]:
            if open:
                self.result += u"^"
        elif tag in [u"p", u"br", u"div"]:
            self.result += u"\n"
        elif tag in [u"ul", u"ol"]:
            if open:
                self.result += u"\n%I"
                self.list_stack.append([tag,0])
            else:
                # Grumble grumble. Bad HTML.
                if len(self.list_stack):
                    self.list_stack.pop()
                self.result += u"%i\n"
        elif tag in [u"li"]:
            if open:
                self.result += u"\n"

                # List item with no start tag, default to ul
                if not len(self.list_stack):
                    self.list_stack.append(["ul",0])

                if self.list_stack[-1][0] == u"ul":
                    self.result += u"\u25CF "
                else:
                    self.list_stack[-1][1] += 1
                    self.result += unicode(self.list_stack[-1][1])+ "."
            else:
                self.result += u"\n"

        elif tag in [u"i", u"small", u"em"]:
            if open:
                self.result += u"%6%B"
            else:
                self.result += u"%b%0"
        elif tag in [u"b", u"strong"]:
            if open:
                self.result += u"%B"
            else:
                self.result += u"%b"

instance = CantoHTML()
def ent_wrapper(match):
    return CantoHTML.convert_entityref(instance,\
         match.groups()[0])

def char_wrapper(match):
    return CantoHTML.convert_charref(instance,\
        match.groups()[0])

def convert(s):
    # We have this try except because under no circumstances
    # should the HTML parser crash the application. Better
    # handling is done per case in the handler itself so that
    # bad HTML doesn't necessarily lead to garbage output.

    try:
        instance.feed(s)
    except:
        pass
    r = instance.result
    l = instance.links
    instance.reset()
    return (r,l)



#!/usr/bin/env python
from __future__ import with_statement # This isn't required in Python 2.6
from distutils.core import setup, Extension
from distutils.command.install_data import install_data

version = ['0','6','13']
man_date = "12 May 2009"

class Canto_install_data(install_data):
    def run(self):
        ret = install_data.run(self)

        install_cmd = self.get_finalized_command('install')
        libdir = install_cmd.install_lib
        mandir = install_cmd.install_data + "/share/man/man1/"

        for source in ["/canto/const.py"]:
            with open(libdir + source, "r+") as f:
                d = f.read().replace("SET_VERSION_TUPLE","(" +\
                        ",".join(version) + ")")
                f.truncate(0)
                f.seek(0)
                f.write(d)

        for manpage in ["canto.1","canto-fetch.1"]:
            with open(mandir + manpage, "r+") as f:
                d = f.read().replace("MAN_VERSION", ".".join(version))
                d = d.replace("MAN_DATE", man_date)
                f.truncate(0)
                f.seek(0)
                f.write(d)

setup(name='Canto',
        version=".".join(version),
        description='An ncurses RSS aggregator.',
        author='Jack Miller',
        author_email='jack@codezen.org',
        url='http://codezen.org/canto',
        license='GPLv2',
        scripts=['bin/canto','bin/canto-fetch'],
        packages=['canto'],
        ext_modules=[Extension('canto.widecurse',\
                sources = ['canto/widecurse.c'], libraries = ['ncursesw'],
                library_dirs=["/usr/local/lib", "/opt/local/lib"],
                include_dirs=["/usr/local/include", "/opt/local/include"])],
        data_files = [("share/man/man1/", ["man/canto.1", "man/canto-fetch.1"])],
        cmdclass={'install_data': Canto_install_data}
)



# -*- coding: utf-8 -*-

#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from const import *
from gui import Gui
from utility import Cycle

import canto_fetch
import utility
import cfg
import tag

import traceback
import signal
import locale
import curses
import getopt
import time
import sys
import os

def print_canto_usage():
    print "USAGE: canto [-hvulaniortDCLF]"
    print "--help      -h        This help."
    print "--version   -v        Print version info."
    print "--update    -u        Fetch updates before running."
    print "--list      -l        List configured feeds."
    print "--checkall  -a        Prints number of new items."
    print "--checknew  -n [feed] Prints number of items that are new in feed."

    print ""
    print "--opml      -o        Convert conf to OPML and print to stdout."
    print "--import    -i [path] Add feeds from OPML file to conf."
    print "--url       -r [url]  Add feed at URL to conf."
    print "--tag       -t [tag]  Set tag (for -r)"
    print ""
    print_common_usage()

def print_fetch_usage():
    print "USAGE: canto-fetch [-hvVfdbDCLF]"
    print "--help       -h       This help."
    print "--version    -v       Print version info."
    print "--verbose    -V       Print extra info while running."
    print "--force      -f       Force update, regardless of timeestamps."
    print "--daemon     -d       Run as a daemon."
    print "--background -b       Background (implies -d)"
    print ""
    print_common_usage()

def print_common_usage():
    print "--dir       -D [path] Set configuration directory. (~/.canto/)"
    print "--conf      -C [path] Set configuration file. (~/.canto/conf)"
    print "--log       -L [path] Set client log file. (~/.canto/log)"
    print "--fdir      -F [path] Set feed directory. (~/.canto/feeds/)"
    print "--sdir      -S [path] Set script directory (~/.canto/scripts/)"

# The Main class encompasses a single instance of Canto or Canto-fetch
# running. It handles arguments and parses the config for both binaries.

class Main():
    def __init__(self):
        # Let locale figure itself out
        locale.setlocale(locale.LC_ALL, "")
        enc = locale.getpreferredencoding()

        # Figure out which binary we are, canto or canto-fetch
        # and determine which arguments we'll accept.

        if sys.argv[0].endswith("canto"):
            shortopts = 'hvulaor:t:i:n:D:C:L:F:S:'
            longopts = ["help","version","update","list","checkall","opml",\
                    "import=","url=","checknew=","dir=",\
                    "conf=","log=","fdir=","sdir=", "tag="]

            iam = "canto"
        elif sys.argv[0].endswith("canto-fetch"):
            shortopts = 'hvVfdbD:C:L:F:S:'
            longopts =   ["help","version","verbose","force","dir=",\
                         "conf=", "log=", "fdir=","sdir=",\
                         "daemon","background"]

            iam = "fetch"
        else:
            print "No idea how you called me..."
            sys.exit(-1)

        # Parse the args.

        try :
            optlist = getopt.getopt(sys.argv[1:],shortopts,longopts)[0]
        except getopt.GetoptError, e:
            print "Error: %s" % e.msg
            sys.exit(-1)

        # Canto and canto-fetch share a certain number of args,
        # mainly the ones dealing with locations of a number of
        # directories and files. These args are all capitalized.

        # Search the args once for changing the root, because
        # the root directory will effect other options.

        for opt, arg in optlist:
            if opt in ["-D", "--dir"]:
                conf_dir = unicode(arg, enc, "ignore")
                break
        else:
            conf_dir = os.getenv("HOME") + "/.canto/"

        if conf_dir[-1] != '/' :
            conf_dir += '/'

        # Now we process the remaining common arguments.

        if iam == "canto":
            log_file = conf_dir + "log"
        else:
            log_file = conf_dir + "fetchlog"

        conf_file = conf_dir + "conf"
        feed_dir = conf_dir + "feeds/"
        script_dir = conf_dir + "scripts/"

        for opt, arg in optlist :
            if opt in ["-C", "--conf"] :
                conf_file = unicode(arg, enc, "ignore")
            elif opt in ["-L","--log"] :
                log_file = unicode(arg, enc, "ignore")
            elif opt in ["-F","--fdir"] :
                feed_dir = unicode(arg, enc, "ignore")
                if feed_dir[-1] != '/' :
                    feed_dir += '/'
            elif opt in ["-S","--sdir"] :
                script_dir = unicode(arg, enc, "ignore")
                if script_dir[-1] != '/' :
                    script_dir += '/'
            elif opt in ["-h","--help"] :
                if iam == "canto":
                    print_canto_usage()
                else:
                    print_fetch_usage()
                sys.exit(0)
            elif opt in ["-v","--version"] :
                print "Canto v %d.%d.%d" % VERSION_TUPLE
                sys.exit(0)

        # Instantiate Cfg() using paths in args.

        try :
            self.cfg = cfg.Cfg(conf_file, log_file, feed_dir, script_dir)
            self.cfg.parse()
        except :
            sys.exit(-1)

        self.cfg.log("Canto v %d.%d.%d" % VERSION_TUPLE, "w")
        self.cfg.log("Time: %s" % time.asctime())
        self.cfg.log("Config parsed successfully.")

        if iam == "fetch":
            # Process canto-fetch specific args.

            daemon = False
            background = False
            for opt, arg in optlist :
                if opt in ["-d","--daemon"]:
                    daemon = True
                if opt in ["-b","--background"]:
                    background = True
                    daemon = True

            # Daemonize the process, which is sorta confusing
            # in this context, because daemonizing is running
            # in the background (separate from the shell)
            # Whereas running as a daemon means looping canto-fetch
            # to avoid the need for a crontab.

            if background:
                utility.daemonize()

            if daemon:
                while 1:
                    canto_fetch.main(self.cfg, optlist)
                    time.sleep(60)
                    oldcfg = self.cfg
                    try :
                        self.cfg = cfg.Cfg(conf_file, log_file, feed_dir,\
                                script_dir)
                        self.cfg.parse()
                    except:
                        self.cfg = oldcfg
            sys.exit(canto_fetch.main(self.cfg, optlist))

        # From this point forward, we are definitely canto,
        # not canto-fetch. Begin processing canto specific args.

        flags = 0 
        feed_ct = None
        opml_file = None
        url = None
        newtag = None

        for opt, arg in optlist :
            if opt in ["-u","--update"] :
                flags |= UPDATE_FIRST
            elif opt in ["-n","--checknew"] :
                flags |= CHECK_NEW
                feed_ct = unicode(arg, enc, "ignore")
            elif opt in ["-a","--checkall"] :
                flags |= CHECK_NEW
            elif opt in ["-l","--list"] :
                flags |= FEED_LIST
            elif opt in ["-o","--opml"] :
                flags |= OUT_OPML
            elif opt in ["-i","--import"] :
                flags |= IN_OPML
                opml_file = unicode(arg, enc, "ignore")
            elif opt in ["-r","--url"] :
                flags |= IN_URL
                url = unicode(arg, enc, "ignore")
            elif opt in ["-t","--tag"] :
                newtag = unicode(arg, enc, "ignore")

        if flags & IN_OPML:
            self.cfg.source_opml(opml_file, append=True)
            print "OPML imported."

        if flags & IN_URL:
            self.cfg.source_url(url, append=True, tag=newtag)
            print "URL added."

        # All import options should terminate.

        if flags & (IN_OPML + IN_URL):
            sys.exit(0)

        # If self.cfg had to generate a config, make sure we
        # update first.

        if self.cfg.no_conf:
            self.cfg.log("Conf was auto-generated, adding -u")
            flags |= UPDATE_FIRST

        if flags & UPDATE_FIRST:
            self.cfg.log("Pausing to update...")
            canto_fetch.main(self.cfg, [], True, True)

        # Detect if there are any new feeds by whether their
        # set path exists. If not, run canto-fetch but don't
        # force it, so canto-fetch intelligently updates.

        for i,f in enumerate(self.cfg.feeds) :
            if not os.path.exists(f.path):
                self.cfg.log("Detected unfetched feed: %s." % f.URL)
                canto_fetch.main(self.cfg, [], True, False)

                #Still no go?
                if not os.path.exists(f.path):
                    self.cfg.log("Failed to fetch %s, removing" % f.URL)
                    self.cfg.feeds[i] = None
                else:
                    self.cfg.log("Fetched.\n")
                break

        # Collapse the feed array, if we had to remove some unfetchables.
        self.cfg.feeds = filter(lambda x: x != None, self.cfg.feeds)

        # The stories array is a "hot" list of stories. After the feed
        # filters and the global filters, stories are shunted into this
        # array, regardless of whether they'll be displayed or anything.
        # It isn't until the Gui requests a refresh or a refresh is
        # generated by an interrupt (SIGWINCH/SIGALRM) that these
        # stories may be displayed.

        self.stories = []

        # Force an update from disk
        self.cfg.log("Populating feeds...")
        for f in self.cfg.feeds:
            try:
                f.time = 1
                f.tick()

            # A KeyError is raised by tick() if canto_version on disk
            # doesn't exist or doesn't match the current info.

            except KeyError:
                self.cfg.log("Detected old feed data, forcing update")
                canto_fetch.main(self.cfg, [], True, True)
                f.time = 1
                f.tick()

        base_tags = {}
        for f in [x for x in self.cfg.feeds if x.base_set]:
            otag = f.tags[0]
            if f.tags[0] in base_tags:
                base_tags[otag] += 1
                while f.tags[0] + (" (%d)" % base_tags[otag]) in base_tags:
                    base_tags[otag] += 1
                f.tags[0] += " (%d)" % base_tags[otag]
                for s in f:
                    s.tagwrap(otag, -1)
                    s.tagwrap(f.tags[0], 1)
            else:
                base_tags[f.tags[0]] = 1
            self.filter_extend(f)

        # Print out a feed list, bail
        if flags & FEED_LIST:
            for f in self.cfg.feeds:
                print f.tags[0].encode(enc, "ignore")
            sys.exit(0)

        # This could probably be done better, or more officially
        # with some XML library, but for now, print is working
        # fairly well.

        if flags & OUT_OPML:
            self.cfg.log("Outputting OPML")
            print """<opml version="1.0">"""
            print """<body>"""
            for feed in self.cfg.feeds:
                if "atom" in feed.ufp["version"]:
                    t = "pie"
                else:
                    t = "rss"

                print """\t<outline text="%s" xmlUrl="%s" type="%s" />""" %\
                        (feed.tags[0].encode(enc, "ignore"), feed.URL, t)

            print """</body>"""
            print """</opml>"""
            sys.exit(0)

        # Handle -a/-n flags (print number of new items)

        if flags & CHECK_NEW:
            if not feed_ct:
                feed_ct = "*"

            # We get counts by using the Tag() class directly.
            check_tag = tag.Tag(self.cfg, Cycle([[None]]),\
                    Cycle([None]), feed_ct)
            check_tag.extend(self.stories)
            print check_tag.unread
            sys.exit(0)

        # At this point we know that we're going to actually launch
        # the client, so we fire up ncurses and add the screen
        # information to our Cfg().

        self.cfg.stdscr = curses.initscr()

        # curs_set can return ERR, we shouldn't care
        try:
                curses.curs_set(0)
        except:
                pass

        # if any of these mess up though, the rest of the
        # the operation is suspect, so die.
        try:
                curses.noecho()
                curses.start_color()
                curses.halfdelay(1)
                curses.use_default_colors()
        except:
                self.cfg.log("Unable to init curses, bailing")
                self.done()

        self.resize = 0
        self.alarmed = 0

        self.cfg.height, self.cfg.width = self.cfg.stdscr.getmaxyx()

        # Init colors
        for i in range(8) :
            f = utility.convcolor(self.cfg.colors[i][0])
            b = utility.convcolor(self.cfg.colors[i][1])
            curses.init_pair(i + 1, f, b)

        self.cfg.log("Curses initialized.")
    
        # Instantiate the base Gui class
        self.cfg.validate_tags()
        Gui(self.cfg, self.stories, self.cfg.tags.cur(), \
                self.push_handler, self.pop_handler)

        self.cfg.log("GUI initialized.")

        # Signal handling
        signal.signal(signal.SIGWINCH, self.winch)
        signal.signal(signal.SIGALRM, self.alarm)
        signal.signal(signal.SIGCHLD, self.chld)
        signal.signal(signal.SIGINT, self.done)

        signal.alarm(1)
        self.ticks = 60

        self.cfg.log("Signals set.")
        self.estring = None

        try:
            # Initial draw of the screen
            self.refresh()

            # Main program loop, terminated when all handlers have
            # deregistered / exited.

            self.cfg.log("Beginning main loop.")

            while 1:
                if not len(self.cfg.key_handlers):
                    break

                t = None

                if self.cfg.wait_for_pid:
                    signal.pause()

                k = self.cfg.stdscr.getch()

                # KEY_RESIZE is the only key not propagated, to
                # keep users from rebinding it and crashing.

                if k == curses.KEY_RESIZE or self.resize:
                    self.resize = 0
                    self.refresh()
                    continue

                # Tick when SIGALRM is received.

                if self.alarmed:
                    self.alarmed = 0
                    self.tick()

                # Handle Meta pairs
                elif k == 195:
                    k2 = self.cfg.stdscr.getch()
                    if k2 >= 64:
                        t = (k2 - 64, 1)
                    else:
                        t = (k, 0)

                # Just a normal key-press
                elif k != -1:
                    t = (k, 0)

                if hasattr(self.cfg.key_handlers[self.cfg.cur_kh], "keys"):
                    if t in self.cfg.key_handlers[self.cfg.cur_kh].keys:
                        actl = self.cfg.key_handlers[self.cfg.cur_kh].keys[t]
                    else:
                        actl = []
                elif t:
                    actl = [t]
                else:
                    actl = []

                for a in actl:
                    if not len(self.cfg.key_handlers):
                        self.done()
                    r = self.cfg.key_handlers[self.cfg.cur_kh].action(a)
                    if r == REFRESH_ALL:
                        self.refresh()
                    elif r == ALARM:
                        self.ticks = 1
                        self.tick()
                    elif r == REDRAW_ALL:
                        for k in self.cfg.key_handlers:
                            k.draw_elements()
                    elif r == WINDOW_SWITCH and len(self.cfg.key_handlers) >= 2:
                        oldcur = self.cfg.key_handlers[self.cfg.cur_kh]
                        if self.cfg.cur_kh == len(self.cfg.key_handlers) - 1:
                            self.cfg.cur_kh = 0
                        else:
                            self.cfg.cur_kh += 1
                        self.update_focus()
                        oldcur.draw_elements()
                        self.cfg.key_handlers[self.cfg.cur_kh].draw_elements()
        except Exception:
            # Catch all _non-exit_ exceptions.
            # -> No bug-report message on things like SystemExit.
            self.estring = traceback.format_exc()
        except KeyboardInterrupt:
            pass

        self.done()

    def done(self, a=None, b=None):
        # Unset signals.
        for s in [signal.SIGALRM, signal.SIGWINCH,
                signal.SIGCHLD, signal.SIGINT]:
            signal.signal(s, signal.SIG_IGN)

        # Kill the message log
        self.cfg.msg = None

        # Kill curses
        curses.endwin()

        self.cfg.log("Curses done.")

        if self.estring:
            self.cfg.log("\nEXCEPTION:")
            self.cfg.log(self.estring)
            print "Canto exited on an exception.\n"
            print self.estring
            print "Please report this bug. Send your logfile " +\
                "(%s) to jack@codezen.org" % self.cfg.log_file

        # Make sure we leave the on-disk presence constant
        for feed in self.cfg.feeds:
            while feed.changed:
                feed.todisk()

        self.cfg.log("Flushed to disk.")
        sys.exit(0)

    def chld(self, a=None, b=None):
        try:
            pid,none = os.wait()
        except:
            return
        if self.cfg.wait_for_pid == pid:
            self.cfg.wait_for_pid = 0
            signal.signal(signal.SIGALRM, self.alarm)
            signal.signal(signal.SIGWINCH, self.winch)
            self.resize = 1

    # The reason KEY_RESIZE is used is that it's unsafe to 
    # do much of anything but set a flag in a signal handler,
    # because data structures could be in half-built states,
    # etc. I'm not sure if Python works around that, but the
    # C programmer in me won't allow me to do it and OpenBSD
    # doesn't even support SIGWINCH, so I won't even count
    # on it.

    def winch(self, a=None, b=None):
        self.resize = 1

    # Similarly, alarm is called every minute, and just sets a flag.

    def alarm(self, a=None, b=None):
        self.alarmed = 1

    def tick(self):
        self.ticks -= 1
        if self.ticks <= 0:
            self.stories = []
            for f in self.cfg.feeds:
                f.tick()
                self.filter_extend(f)
    
            for h in self.cfg.key_handlers:
                h.alarm(self.stories)
            self.cfg.key_handlers[self.cfg.cur_kh].refresh()
            self.ticks = 60

        self.cfg.msg_tick -= 1
        if self.cfg.msg_tick == 0:
            self.cfg.message(self.cfg.status(self.cfg), 1)

        signal.alarm(1)

    # Refresh should only be called initially, if we have a 
    # resize event, or if it's possible that the terminal has
    # been resized in our absence (eg. we've just gotten
    # control back from a text browser).

    # Refresh generally causes gui objects to rebuild window
    # objects and redraw the screen, causing flicker.

    def refresh(self):
        curses.endwin()
        self.cfg.stdscr.touchwin()
        self.cfg.stdscr.refresh()
        self.cfg.height, self.cfg.width = self.cfg.stdscr.getmaxyx()

        if self.cfg.resize_hook:
            self.cfg.resize_hook(self.cfg)

        # Make sure we've got a minimum for columns and reader_lines
        self.cfg.columns = max(self.cfg.columns, 1)
        self.cfg.reader_lines = max(self.cfg.reader_lines, 3)

        self.cfg.gui_height = self.cfg.height - self.cfg.msg_height
        self.cfg.gui_width = self.cfg.width

        # This logic could be cleaned up...

        if self.cfg.reader_orientation == "top":
            self.cfg.gui_height -= self.cfg.reader_lines
            self.cfg.gui_top = self.cfg.reader_lines
        elif self.cfg.reader_orientation == "bottom":
            self.cfg.gui_height -= self.cfg.reader_lines
        elif self.cfg.reader_orientation == "left":
            self.cfg.gui_width -= self.cfg.reader_lines
            self.cfg.gui_right = self.cfg.reader_lines
        elif self.cfg.reader_orientation == "right":
            self.cfg.gui_width -= self.cfg.reader_lines
            
        self.cfg.msg = curses.newwin(self.cfg.msg_height,\
                self.cfg.width, self.cfg.height - self.cfg.msg_height, 0)
        self.cfg.msg.bkgdset(curses.color_pair(1))
        self.cfg.msg.erase()
        self.cfg.msg.refresh()

        self.cfg.stdscr.keypad(1)

        for g in self.cfg.key_handlers :
            g.refresh()

    # These two functions are known as register() and deregister()
    # to the gui objects, and let the Main() class know when a gui
    # object should start or stop receiving input.

    def update_focus(self):
        if len(self.cfg.key_handlers):
            for h in self.cfg.key_handlers:
                h.focus = 0
            self.cfg.key_handlers[self.cfg.cur_kh].focus = 1

    def push_handler(self, handler):
        self.cfg.key_handlers.append(handler)
        self.cfg.cur_kh = len(self.cfg.key_handlers) - 1
        self.update_focus()

    def pop_handler(self):
        self.cfg.key_handlers.remove(self.cfg.key_handlers[self.cfg.cur_kh])
        self.cfg.cur_kh = max(self.cfg.cur_kh - 1, 0)
        self.update_focus()
        if len(self.cfg.key_handlers):
           for h in self.cfg.key_handlers:
               h.refresh()

    # Filter extend extends self.stories with items passing through
    # the global filter. The Feed() objects are never changed.

    def filter_extend(self, t):
        filt = self.cfg.filters.cur()
        if filt:
            self.stories.extend(filter(lambda x: filt(t,x), t))
        else:
            self.stories.extend(t)




# -*- coding: utf-8 -*-

#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.


from widecurse import core, tlen
import canto_html

import locale
import re

class Renderer :
    def __init__(self):
        self.prefcode = locale.getpreferredencoding()

        self.story_rgx = [
            # Eliminate extraneous HTML
            (re.compile(u"<.*?>"), u""),
            (re.compile(u"&(\w{1,8});"), canto_html.ent_wrapper),
            (re.compile(u"&#([xX]?[0-9a-fA-F]+)[^0-9a-fA-F]"),
                canto_html.char_wrapper)
            ]

        self.reader_pre_rgx = []

        self.reader_post_rgx = [
            (re.compile(u"[\\\"](.*?)[\\\"]"), u"%5\"\\1\"%0"),
            ]

        self.bq = u"%B%1│%0%b "
        self.bq_on = 0

        self.indent = u"  "
        self.in_on = 0

    def tag_head(self, dict):
        t = u"%1" + dict["tag"].tag + u" [%2" + unicode(dict["tag"].unread)\
                + u"%0]%0"
        if dict["tag"].collapsed:
            if dict["tag"][0].selected():
                return [(u"%C%B%1 > " + t + u"", u" ", u" "),(u" ",u" ",u" ")]
            else:
                return [(u"%C%B   " + t + u"",u" ", u" "),(u" ",u" ",u" ")]

        return [(u"%B   " + t, u" ", u""),(u"%1┌", u"─", u"┐%C%0")]

    def tag_foot(self, dict):
        return [(u"%1%B└", u"─", u"┘%C%0")]

    def firsts(self, dict):
        base = u"%C%1%B│%b%0 "
    
        if dict["story"].selected() :
            if dict["cfg"].key_handlers[0].focus:
                base += u"%1%B>%b%0 "
            else:
                base += u"%1%B_%b%0 "
        else:
            base += u"  "

        if dict["story"].marked():
            base += u"%1%B"
        else:
            if dict["story"].wasread():
                base += u"%3"
            else:
                base += u"%2%B"

        return (base, u" ", u" %1%B│%b%0")

    def mids(self, dict):
        return (u"%1%B│%b%0      ", u" ", u" %1%B│%b%0")

    def ends(self, dict):
        return (u"%1%B│%b%0      ", u" ", u" %1%B│%b%0")

    def reader_head(self, dict):
        title = self.do_regex(dict["story"]["title"], self.story_rgx)
        return [(u"%1%B" + title, u" ", u" "),(u"%1┌",u"─",u"┐%C")]

    def reader_foot(self, dict):
        return [(u"%B└", u"─", u"┘%C")]

    def reader_link(self, idx, link):
        if link[2] == "link":
            color = u"%4"
        elif link[2] == "image":
            color = u"%7"
        else:
            color = u"%8"

        return color + u"[" + unicode(idx) + u"] " + \
                link[0] + u"%1 - " + link[1]

    def rfirsts(self, dict):
        return (u"%1%B│%b%0 ", u" ", u" %1%B│%b%0")

    def rmids(self, dict):
        return (u"%1%B│%b%0 ", u" ", u" %1%B│%b%0")
    
    def rends(self, dict):
        return (u"%1%B│%b%0 ", u" ", u" %1%B│%b%0")

    def __window(self, row, height, window_list):
        if height != -1:
            winidx, winrow = divmod(row, height)
            try :
                window = window_list[winidx]
            except IndexError:
                window = None
            return (window, winrow)
        else:
            return (window_list[0], row)

    def core_wrap(self, window, winrow, width, s, rep, end):
        ret = core(window, winrow, 0, width,
                s.encode(self.prefcode, 'replace'),
                rep.encode(self.prefcode, 'replace'),
                end.encode(self.prefcode, 'replace'))
        if ret:
            ret = unicode(ret, self.prefcode)
        return ret

    def tlen_wrap(self, s):
        return tlen(s.encode(self.prefcode, 'replace'))

    def simple_out(self, list, row, height, width, window_list):
        line = 0
        for s,rep,end in list:
            while s:
                 window, winrow = self.__window(row + line, height, window_list)
                 s = self.core_wrap(window, winrow, width, s, rep, end)
                 line += 1

        return row + line
               
    def out(self, list, row, height, width, window_list):
        line = 0
        for s, l in list:
            if s:

                # Handle block level style, not covered in widecurse.
                # This is broken into three sections so that styles
                # can be applied to a single line and applied in between.

                # Note, as with any unknown % escape, these will be
                # totally ignored in the middle of a line.

                # Toggle on based on start of line
                while s[:2] in [u"%Q",u"%I"]:
                    if s.startswith(u"%Q"):
                        self.bq_on += 1
                    else:
                        self.in_on += 1
                    s = s[2:]

                # Add decorations to firsts,mids,lasts
                if self.bq_on:
                    l = [(e[0] + self.bq * self.bq_on,\
                            e[1],e[2]) for e in l]
                if self.in_on:
                    l = [(e[0] + self.indent * self.in_on,\
                            e[1],e[2]) for e in l]
               
                # Toggle off based on end of line
                while s[-2:] in [u"%q",u"%i"]:
                    if s.endswith(u"%q"):
                        self.bq_on -= 1
                    else:
                        self.in_on -= 1
                    s = s[:-2]

            while s :
                window, winrow = self.__window(row + line, height, window_list)

                # First line, obviously use first line caps.
                if line == 0:
                    start, rep, end = l[0]
                
                # If line > 1 and we've got more than could be handled
                # with end_caps, use mid_caps

                elif self.tlen_wrap(s) > (width - (self.tlen_wrap(l[2][2]))):
                    start, rep, end = l[1]

                # Otherwise, use end_caps

                else:
                    start, rep, end = l[2]

                t = s
                s = self.core_wrap(window, winrow, width, start + s, rep, end)

                # Detect an infinite loop caused by start, and canto
                # trying to be smart about wrapping =).

                if s == t:
                    s = self.core_wrap(window, winrow, width, s, u" ",u"")
                line += 1

        return row + line

    def do_regex(self, target, rlist):
        s = target
        for rgx,rep in rlist:
            s = rgx.sub(rep,s)
        return s
    
    def story(self, cfg, tag, story, row, height, width, window_list):
        title = self.do_regex(story["title"], self.story_rgx)
        title = title.lstrip().rstrip()

        d = {"tag": tag, "story" : story, "cfg" : cfg }

        if story.idx == 0:
            row = self.simple_out(self.tag_head(d),\
                row, height, width, window_list)

        if not tag.collapsed:
            row = self.out([[title, (self.firsts(d), self.mids(d), \
                    self.ends(d))]],
                    row, height, width, window_list)
            
            if story.last:
                row = self.simple_out(self.tag_foot(d),\
                    row, height, width, window_list)
    
        return row

    def reader(self, cfg, story, width, show_links, window):
        s = story.get_text()

        enc_links = []
        if "enclosures" in story:
            for e in story["enclosures"]:
                enc_links.append((u"[%s]" % e["type"],
                        e["href"], "link"))

        d = {"story" : story, "cfg" : cfg }

        s = self.do_regex(s, self.reader_pre_rgx)
        s,links = canto_html.convert(s)
        s = self.do_regex(s, self.reader_post_rgx)

        links = [(u"main link", story["link"], "link")] + links
        links += enc_links

        l = s.split("\n")
        if show_links:
            l.append(" ")
            for idx,link in enumerate(links):
                l.append(self.reader_link(idx, link))

        row = self.simple_out(self.reader_head(d), 0, -1, width, [window])
        row = self.out([[x, (self.rfirsts(d), self.rmids(d),
            self.rends(d))] for x in l], row, -1, width, [window])
        row = self.simple_out(self.reader_foot(d), row, -1, width, [window])
        return row, links

    def status(self, bar, height, width, str):
        self.simple_out([(str, u" ", u"")], 0, height, width, [bar])



# -*- coding: utf-8 -*-

#Canto - ncurses RSS reader
#   Copyright (C) 2008 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from utility import Cycle
import story

class Tag(list):
    def __init__(self, cfg, sorts = [[None]], filters = [None], c = "*"):
        list.__init__(self)
        self.cfg = cfg
        self.tag = c
        self.collapsed = 0
        self.start = 0
        self.read = 0
        self.unread = 0
        self.last_iter = []

        self.filters = filters
        self.sorts = sorts

    def __eq__(self, other):
        return self.tag == other.tag

    def __str__(self):
        return self.tag

    def search_stories(self, story):
        for i in range(len(self)) :
            if self[i]["id"] == story["id"]:
                return i
        return -1

    def all_read(self):
        for s in self :
            s.read()
        self.unread = 0
        self.read = len(self)

    def all_unread(self):
        for s in self :
            s.unread()
        self.read = 0
        self.unread = len(self)

    def set_read(self, idx):
        if not self[idx].wasread():
            self[idx].read()
            self.unread -= 1
            self.read += 1

    def set_unread(self, idx):
        if self[idx].wasread():
            self[idx].unread()
            self.unread += 1
            self.read -= 1

    def extend(self, iter):
        self.last_iter = iter
        matched_tag = [s for s in iter if self.tag in s["canto_state"]]

        filt = self.filters.cur()
        if filt:
            list.extend(self, filter(lambda x: filt(self, x), matched_tag))
        else:
            list.extend(self, matched_tag)

        empty = 0
        if filt and not len(self):
            d = { "title" : "No unfiltered items.",
                  "description" : "You've filtered out everything!",
                  "canto_state" : [self.tag, "unread"],
                  "id" : None
                }

            stub = story.Story(d , None, self.cfg.default_renderer)
            self.append(stub)
            empty = 1
        else:
            if not hasattr(self.sorts.cur(), "__iter__"):
                dosorts = [self.sorts.cur()]
            else:
                dosorts = self.sorts.cur()

            for s in dosorts:
                if s:
                    list.sort(self, s)

        if empty:
            self.read = 0
            self.unread = 0
            self[0].idx = 0
            self[0].last = 1
        else:
            lt = len(self)
            for i in range(lt):
                self[i].idx = i
                self[i].last = 0
            if lt:
                self[-1].last = 1

            self.read = len(filter(lambda x : x.wasread(), self))
            self.unread = len(self) - self.read

    def clear(self):
        del self[:]
